from collections import deque

from cpython.datetime cimport datetime
from cpython.datetime cimport timedelta
from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.core.datetime cimport is_datetime_utc
from nautilus_trader.core.datetime cimport unix_nanos_to_dt
from nautilus_trader.indicators.base.indicator cimport Indicator
from nautilus_trader.model.data.bar cimport Bar
from nautilus_trader.model.identifiers cimport InstrumentId
from nautilus_trader.model.objects cimport Price


cdef class BarMinMax(Indicator):
    """
    Given a historic lookback window of high/low prices and volume, keep a running
    computation of the min/max values of the high/low prices and volume within the window.
    """

    def __init__(self, timedelta lookback not None):
        """
        Initialize a new instance of the ``BarMinMax`` class.

        Parameters
        ----------
        lookback : timedelta
            The look back duration in time.

        """
        super().__init__(params=[lookback])

        self.lookback = lookback

        self.prices = WindowedMinMaxPrices(lookback)
        self.volumes = WindowedMinMaxPrices(lookback)

    cpdef void handle_bar(self, Bar bar) except *:
        self.prices.add_price(unix_nanos_to_dt(nanos=bar.ts_event), bar.high)
        self.prices.add_price(unix_nanos_to_dt(nanos=bar.ts_event), bar.low)

        cdef Price volume_as_price = Price(bar.volume._value, bar.volume.precision)
        self.volumes.add_price(unix_nanos_to_dt(nanos=bar.ts_event), volume_as_price)

        # Mark as having input and initialized
        self._set_has_inputs(True)
        self._set_initialized(True)

    cpdef void _reset(self) except *:
        # Reset the windows
        self.prices.reset()
        self.volumes.reset()

cdef class WindowedMinMaxPrices:
    """
    Over the course of a defined lookback window, efficiently keep track
    of the min/max values currently in the window.
    """

    def __init__(self, timedelta lookback not None):
        """
        Initialize a new instance of the ``WindowedMinMaxPrices`` class.
        Parameters
        ----------
        lookback : timedelta
            The look back duration in time.
        """
        self.lookback = lookback

        # Initialize the deques
        self._min_prices = deque()
        self._max_prices = deque()

        # Set the min/max marks as None until we have data
        self.min_price = None
        self.max_price = None

    cpdef void add_price(self, datetime ts, Price price) except *:
        """
        Given a price at a UTC timestamp, insert it into the structures and
        update our running min/max values.
        Parameters
        ----------
        ts : datetime
            The timestamp for the price.
        price : Price
            The price to add.
        """
        Condition.true(is_datetime_utc(ts), "ts was not tz-aware UTC")

        # Expire old prices
        cdef datetime cutoff = ts - self.lookback
        self._expire_stale_prices_by_cutoff(self._min_prices, cutoff)
        self._expire_stale_prices_by_cutoff(self._max_prices, cutoff)

        # Append to the min/max structures
        self._add_min_price(ts, price)
        self._add_max_price(ts, price)

        # Pull out the min/max
        self.min_price = min([p[1] for p in self._min_prices])
        self.max_price = max([p[1] for p in self._max_prices])

    cpdef void reset(self) except *:
        """
        Reset the indicator.
        All stateful fields are reset to their initial value.
        """
        # Set the min/max marks as None until we have data
        self.min_price = None
        self.max_price = None

        # Clear the deques
        self._min_prices.clear()
        self._max_prices.clear()

    cdef void _expire_stale_prices_by_cutoff(
            self,
            ts_prices,
            datetime cutoff
    ) except *:
        """Drop items that are older than the cutoff"""
        while ts_prices and ts_prices[0][0] < cutoff:
            ts_prices.popleft()

    cdef void _add_min_price(self, datetime ts, Price price) except *:
        """Handle appending to the min deque"""
        # Pop front elements that are less than or equal (since we want the max ask)
        while self._min_prices and self._min_prices[-1][1] >= price:
            self._min_prices.pop()

        # Pop back elements that are less than or equal to the new ask
        while self._min_prices and self._min_prices[0][1] >= price:
            self._min_prices.popleft()

        self._min_prices.append((ts, price))

    cdef void _add_max_price(self, datetime ts, Price price) except *:
        """Handle appending to the max deque"""
        # Pop front elements that are less than or equal (since we want the max bid)
        while self._max_prices and self._max_prices[-1][1] <= price:
            self._max_prices.pop()

        # Pop back elements that are less than or equal to the new bid
        while self._max_prices and self._max_prices[0][1] <= price:
            self._max_prices.popleft()

        self._max_prices.append((ts, price))



from decimal import Decimal
from typing import Any, Dict

from nautilus_trader.model.data.bar import Bar
from nautilus_trader.model.data.bar import BarType
from nautilus_trader.model.data.ticker import Ticker
from nautilus_trader.model.identifiers import InstrumentId
from nautilus_trader.model.objects import Price
from nautilus_trader.model.objects import Quantity


class ZbTicker(Ticker):
    def __init__(
        self,
        instrument_id: InstrumentId,
        price_change: Decimal,
        last_price: Decimal,
        open_price: Decimal,
        high_price: Decimal,
        low_price: Decimal,
        volume: Decimal,
        last_id: int,
        ts_event: int,
        ts_init: int,
    ):
        """
        Initialize a new instance of the ``ZbTicker`` class.

        Parameters
        ----------
        instrument_id : InstrumentId
            The instrument ID.
        price_change : Decimal
            The price change.
        last_price : Decimal
            The last price.
        open_price : Decimal
            The open price.
        high_price : Decimal
            The high price.
        low_price : Decimal
            The low price.
        volume : Decimal
            The volume.
        last_id : int
            The last trade time
        ts_event : int64
            The UNIX timestamp (nanoseconds) when the ticker event occurred.
        ts_init : int64
            The UNIX timestamp (nanoseconds) when the object was initialized.

        """
        super().__init__(
            instrument_id=instrument_id,
            ts_event=ts_event,
            ts_init=ts_init,
        )

        self.price_change = price_change
        self.last_price = last_price
        self.open_price = open_price
        self.high_price = high_price
        self.low_price = low_price
        self.volume = volume
        self.last_id = last_id

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"instrument_id={self.instrument_id.value}, "
            f"price_change={self.price_change}, "
            f"last_price={self.last_price}, "
            f"open_price={self.open_price}, "
            f"high_price={self.high_price}, "
            f"low_price={self.low_price}, "
            f"volume={self.volume}, "
            f"last_id={self.last_id}, "
            f"ts_event={self.ts_event},"
            f"ts_init={self.ts_init})"
        )

    @staticmethod
    def from_dict(values: Dict[str, Any]) -> "ZbTicker":
        """
        Return a `Zb` ticker parsed from the given values.

        Parameters
        ----------
        values : dict[str, Any]
            The values for initialization.

        Returns
        -------
        ZbTicker

        """
        return ZbTicker(
            instrument_id=InstrumentId.from_str(values["instrument_id"]),
            price_change=Decimal(values["price_change"]),
            last_price=Decimal(values["last_price"]),
            open_price=Decimal(values["open_price"]),
            high_price=Decimal(values["high_price"]),
            low_price=Decimal(values["low_price"]),
            volume=Decimal(values["volume"]),
            last_id=values["last_id"],
            ts_event=values["ts_event"],
            ts_init=values["ts_init"],
        )

    @staticmethod
    def to_dict(obj: "ZbTicker") -> Dict[str, Any]:
        """
        Return a dictionary representation of this object.

        Returns
        -------
        dict[str, Any]

        """
        return {
            "type": type(obj).__name__,
            "instrument_id": obj.instrument_id.value,
            "price_change": str(obj.price_change),
            "last_price": str(obj.last_price),
            "open_price": str(obj.open_price),
            "high_price": str(obj.high_price),
            "low_price": str(obj.low_price),
            "volume": str(obj.volume),
            "last_id": obj.last_id,
            "ts_event": obj.ts_event,
            "ts_init": obj.ts_init,
        }


class ZbBar(Bar):
    def __init__(
        self,
        bar_type: BarType,
        open: Price,
        high: Price,
        low: Price,
        close: Price,
        volume: Quantity,
        ts_event: int,
        ts_init: int,
    ):
        """
        Initialize a new instance of the ``ZbBar`` class.

        Parameters
        ----------
        bar_type : BarType
            The bar type for this bar.
        open : Price
            The bars open price.
        high : Price
            The bars high price.
        low : Price
            The bars low price.
        close : Price
            The bars close price.
        volume : Quantity
            The bars volume.
        ts_event : int64
            The UNIX timestamp (nanoseconds) when the data event occurred.
        ts_init: int64
            The UNIX timestamp (nanoseconds) when the data object was initialized.

        """
        super().__init__(
            bar_type=bar_type,
            open=open,
            high=high,
            low=low,
            close=close,
            volume=volume,
            ts_event=ts_event,
            ts_init=ts_init,
        )

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"bar_type={self.type}, "
            f"open={self.open}, "
            f"high={self.high}, "
            f"low={self.low}, "
            f"close={self.close}, "
            f"volume={self.volume}, "
            f"ts_event={self.ts_event},"
            f"ts_init={self.ts_init})"
        )

    @staticmethod
    def from_dict(values: Dict[str, Any]) -> "ZbBar":
        """
        Return a `Zb` bar parsed from the given values.

        Parameters
        ----------
        values : dict[str, Any]
            The values for initialization.

        Returns
        -------
        ZbBar

        """
        return ZbBar(
            bar_type=BarType.from_str(values["bar_type"]),
            open=Price.from_str(values["open"]),
            high=Price.from_str(values["high"]),
            low=Price.from_str(values["low"]),
            close=Price.from_str(values["close"]),
            volume=Quantity.from_str(values["volume"]),
            ts_event=values["ts_event"],
            ts_init=values["ts_init"],
        )

    @staticmethod
    def to_dict(obj: "ZbBar") -> Dict[str, Any]:
        """
        Return a dictionary representation of this object.

        Returns
        -------
        dict[str, Any]

        """
        return {
            "type": type(obj).__name__,
            "bar_type": str(obj.type),
            "open": str(obj.open),
            "high": str(obj.high),
            "low": str(obj.low),
            "close": str(obj.close),
            "volume": str(obj.volume),
            "ts_event": obj.ts_event,
            "ts_init": obj.ts_init,
        }



from cpython.datetime cimport datetime

from collections import deque

import numpy as np

from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.core.stats cimport fast_mean
from nautilus_trader.indicators.base.indicator cimport Indicator
from nautilus_trader.model.data.bar cimport Bar


cdef class Avl(Indicator):
    """
    An indicator which calculates the volume weighted average price for the day.
    """

    def __init__(self, int period):
        """
        Initialize a new instance of the ``Avl`` class.
        """
        Condition.positive_int(period, "period")
        super().__init__(params=[period])

        self._value_inputs = deque(maxlen=period)
        self._volume_inputs = deque(maxlen=period)
        self.value = 0
        self.count = 0
        self.period = period

    cpdef void handle_bar(self, Bar bar) except *:
        """
        Update the indicator with the given bar.

        Parameters
        ----------
        bar : Bar
            The update bar.

        """
        Condition.not_none(bar, "bar")

        prices = np.asarray(
            [
                bar.open.as_double(),
                bar.high.as_double(),
                bar.low.as_double(),
                bar.close.as_double(),
            ], dtype=np.float64)

        cdef double value = fast_mean(prices) * bar.volume.as_double()

        self.update_raw(
            value,
            bar.volume.as_double(),
        )

    cpdef void update_raw(
        self,
        double value,
        double volume,
    ) except *:
        """
        Update the indicator with the given raw values.

        Parameters
        ----------
        price : double
            The update price.
        volume : double
            The update volume.
        timestamp : datetime
            The current timestamp.

        """

        self._increment_count()
        self._value_inputs.append(value)
        self._volume_inputs.append(volume)

        self.value = sum(self._value_inputs) / sum(self._volume_inputs) if sum(self._volume_inputs) else 0

    cpdef void _reset(self) except *:
        self._inputs.clear()
        self.value = 0
        self.count = 0

    cpdef void _increment_count(self) except *:
        self.count += 1

        # Initialization logic
        if not self.initialized:
            self._set_has_inputs(True)
            if self.count >= self.period:
                self._set_initialized(True)



# flake8: noqa
# mypy: ignore-errors
import os

import nautilus_trader
from nautilus_trader.adapters.binance.http import client as http_client
from nautilus_trader.adapters.binance.http import error
from nautilus_trader.adapters.binance.websocket import client as ws_client

from nacre.adapters.binance.parsing import execution
from nacre.network.http import HttpClient
from nacre.network.websocket import WebSocketClient


def __repr__(self) -> str:
    return f"{type(self).__name__}(" f"status={self.status}, " f"message={self.message})"


error.BinanceError.__repr__ = __repr__

http_client.BinanceHttpClient = type(
    "BinanceHttpClient", (HttpClient,), dict(http_client.BinanceHttpClient.__dict__)
)
NAUTILUS_VERSION = nautilus_trader.__version__


# Patch BinanceHttpClient using nacre.network.http.HttpClient
def http_init(
    self,
    loop,
    clock,
    logger,
    key=None,
    secret=None,
    base_url=None,
    timeout=None,
    show_limit_usage=False,
    proxy=None,
):
    super(http_client.BinanceHttpClient, self).__init__(
        loop=loop,
        logger=logger,
    )
    self._clock = clock
    self._key = key
    self._secret = secret
    self._base_url = base_url or self.BASE_URL
    self._show_limit_usage = show_limit_usage
    self._proxies = None
    self._headers = {
        "Content-Type": "application/json;charset=utf-8",
        "User-Agent": "nautilus-trader/" + NAUTILUS_VERSION,
        "X-MBX-APIKEY": key,
    }
    self._proxy = proxy if proxy else os.environ.get("PROXY")

    if timeout is not None:
        self._headers["timeout"] = timeout


async def request(self, *args, **kwargs):
    if "proxy" not in kwargs and self._proxy:
        kwargs["proxy"] = self._proxy
    return await super(http_client.BinanceHttpClient, self).request(*args, **kwargs)


http_client.BinanceHttpClient.__init__ = http_init
http_client.BinanceHttpClient.request = request

# Patch BinanceWebSocketClient using nacre.network.websocket.WebSocketClient


ws_client.BinanceWebSocketClient = type(
    "BinanceWebSocketClient", (WebSocketClient,), dict(ws_client.BinanceWebSocketClient.__dict__)
)


def ws_init(
    self,
    loop,
    clock,
    logger,
    handler,
    base_url=None,
):
    super(ws_client.BinanceWebSocketClient, self).__init__(
        loop=loop,
        logger=logger,
        handler=handler,
        max_retry_connection=6,
    )

    self._base_url = base_url

    self._clock = clock
    self._streams = []


async def ws_connect(
    self,
    key=None,
    start=True,
    **ws_kwargs,
) -> None:
    if not self._streams:
        raise RuntimeError("No subscriptions for connection.")

    # Always connecting combined streams for consistency
    ws_url = self._base_url + "/stream?streams=" + "/".join(self._streams)
    if key is not None:
        ws_url += f"&listenKey={key}"

    self._log.info(f"Connecting to {ws_url}")
    await super(ws_client.BinanceWebSocketClient, self).connect(
        ws_url=ws_url, start=start, **ws_kwargs
    )


ws_client.BinanceWebSocketClient.__init__ = ws_init
ws_client.BinanceWebSocketClient.connect = ws_connect




# -------------------------------------------------------------------------------------------------
#  Copyright (C) 2015-2022 Nautech Systems Pty Ltd. All rights reserved.
#  https://nautechsystems.io
#
#  Licensed under the GNU Lesser General Public License Version 3.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at https://www.gnu.org/licenses/lgpl-3.0.en.html
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# -------------------------------------------------------------------------------------------------

from libc.stdint cimport int64_t
from nautilus_trader.cache.base cimport CacheFacade
from nautilus_trader.common.clock cimport Clock
from nautilus_trader.common.logging cimport Logger
from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.core.uuid cimport UUID4
from nautilus_trader.data.engine cimport DataEngine
from nautilus_trader.execution.engine cimport ExecutionEngine
from nautilus_trader.model.identifiers cimport TraderId
from nautilus_trader.msgbus.bus cimport MessageBus
from nautilus_trader.portfolio.base cimport PortfolioFacade
from nautilus_trader.risk.engine cimport RiskEngine
from nautilus_trader.system.kernel cimport NautilusKernel as Kernel
from nautilus_trader.trading.trader cimport Trader


cdef class NautilusKernel(Kernel):
    pass



import time
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

from nautilus_trader.common.providers import InstrumentProvider
from nautilus_trader.core.datetime import millis_to_nanos
from nautilus_trader.core.datetime import secs_to_nanos
from nautilus_trader.core.uuid import UUID4
from nautilus_trader.execution.reports import OrderStatusReport
from nautilus_trader.execution.reports import PositionStatusReport
from nautilus_trader.execution.reports import TradeReport
from nautilus_trader.model.currency import Currency
from nautilus_trader.model.data.bar import BarSpecification
from nautilus_trader.model.data.bar import BarType
from nautilus_trader.model.data.base import DataType
from nautilus_trader.model.data.base import GenericData
from nautilus_trader.model.data.tick import TradeTick
from nautilus_trader.model.enums import AggregationSource
from nautilus_trader.model.enums import AggressorSide
from nautilus_trader.model.enums import BarAggregation
from nautilus_trader.model.enums import BookAction
from nautilus_trader.model.enums import BookType
from nautilus_trader.model.enums import CurrencyType
from nautilus_trader.model.enums import LiquiditySide
from nautilus_trader.model.enums import OMSType
from nautilus_trader.model.enums import OrderSide
from nautilus_trader.model.enums import OrderStatus
from nautilus_trader.model.enums import OrderType
from nautilus_trader.model.enums import PositionSide
from nautilus_trader.model.enums import PriceType
from nautilus_trader.model.enums import TimeInForce
from nautilus_trader.model.identifiers import AccountId
from nautilus_trader.model.identifiers import ClientOrderId
from nautilus_trader.model.identifiers import InstrumentId
from nautilus_trader.model.identifiers import PositionId
from nautilus_trader.model.identifiers import Symbol
from nautilus_trader.model.identifiers import TradeId
from nautilus_trader.model.identifiers import Venue
from nautilus_trader.model.identifiers import VenueOrderId
from nautilus_trader.model.instruments.base import Instrument
from nautilus_trader.model.instruments.crypto_perpetual import CryptoPerpetual
from nautilus_trader.model.instruments.currency_pair import CurrencyPair
from nautilus_trader.model.objects import AccountBalance
from nautilus_trader.model.objects import MarginBalance
from nautilus_trader.model.objects import Money
from nautilus_trader.model.objects import Price
from nautilus_trader.model.objects import Quantity
from nautilus_trader.model.orderbook.data import Order
from nautilus_trader.model.orderbook.data import OrderBookDelta
from nautilus_trader.model.orderbook.data import OrderBookDeltas
from nautilus_trader.model.orderbook.data import OrderBookSnapshot
from nautilus_trader.model.orders.limit import LimitOrder
from nautilus_trader.model.orders.market import MarketOrder
from nautilus_trader.model.position import Position

from nacre.adapters.zb.data_types import ZbBar
from nacre.adapters.zb.data_types import ZbTicker
from nacre.model.data.tick import MarkTick


def parse_mark_price(instrument_id: InstrumentId, price: str, ts_init: int) -> MarkTick:
    tick = MarkTick(
        instrument_id=instrument_id,
        price=Price.from_str(price),
        ts_event=ts_init,
        ts_init=ts_init,
    )
    return GenericData(
        data_type=DataType(MarkTick, metadata={"instrument_id": instrument_id}),
        data=tick,
    )


def parse_book_snapshot_ws(
    instrument_id: InstrumentId, msg: Dict, ts_init: int
) -> OrderBookSnapshot:
    ts_event: int = ts_init

    return OrderBookSnapshot(
        instrument_id=instrument_id,
        book_type=BookType.L2_MBP,
        bids=[[float(o[0]), float(o[1])] for o in msg.get("bids", [])],
        asks=[[float(o[0]), float(o[1])] for o in msg.get("asks", [])],
        ts_event=ts_event,
        ts_init=ts_init,
        update_id=int(msg.get("time")),
    )


def parse_diff_depth_stream_ws(
    instrument_id: InstrumentId, msg: Dict, ts_init: int
) -> OrderBookDeltas:
    ts_event: int = ts_init
    update_id: int = int(msg["time"])

    bid_deltas = [
        parse_book_delta_ws(instrument_id, OrderSide.BUY, d, ts_event, ts_init, update_id)
        for d in msg.get("bids", [])
    ]
    ask_deltas = [
        parse_book_delta_ws(instrument_id, OrderSide.SELL, d, ts_event, ts_init, update_id)
        for d in msg.get("asks", [])
    ]

    return OrderBookDeltas(
        instrument_id=instrument_id,
        book_type=BookType.L2_MBP,
        deltas=bid_deltas + ask_deltas,
        ts_event=ts_event,
        ts_init=ts_init,
        update_id=update_id,
    )


def parse_book_delta_ws(
    instrument_id: InstrumentId,
    side: OrderSide,
    delta: Tuple[str, str],
    ts_event: int,
    ts_init: int,
    update_id: int,
) -> OrderBookDelta:
    price = float(delta[0])
    size = float(delta[1])

    order = Order(
        price=price,
        size=size,
        side=side,
    )

    return OrderBookDelta(
        instrument_id=instrument_id,
        book_type=BookType.L2_MBP,
        action=BookAction.UPDATE if size > 0.0 else BookAction.DELETE,
        order=order,
        ts_event=ts_event,
        ts_init=ts_init,
        update_id=update_id,
    )


def parse_spot_ticker_ws(
    instrument_id: InstrumentId, msg: Dict[str, str], ts_init: int
) -> ZbTicker:
    return ZbTicker(
        instrument_id=instrument_id,
        price_change=Decimal(msg["riseRate"]),
        last_price=Decimal(msg["last"]),
        open_price=Decimal(msg["open"]),
        high_price=Decimal(msg["high"]),
        low_price=Decimal(msg["low"]),
        volume=Decimal(msg["vol"]),
        last_id=0,  # not return in spot
        # ts_event=secs_to_nanos(float(msg[6])-1),  # zb time not accurate
        ts_event=ts_init,
        ts_init=ts_init,
    )


def parse_ticker_ws(instrument_id: InstrumentId, msg: List, ts_init: int) -> ZbTicker:
    return ZbTicker(
        instrument_id=instrument_id,
        price_change=Decimal(msg[5]),
        last_price=Decimal(msg[3]),
        open_price=Decimal(msg[0]),
        high_price=Decimal(msg[1]),
        low_price=Decimal(msg[2]),
        volume=Decimal(msg[4]),
        last_id=msg[6],
        # ts_event=secs_to_nanos(float(msg[6])-1),  # zb time not accurate
        ts_event=ts_init,
        ts_init=ts_init,
    )


def parse_trade_tick(instrument_id: InstrumentId, msg: List, ts_init: int) -> TradeTick:
    return TradeTick(
        instrument_id=instrument_id,
        price=Price.from_str(str(msg[0])),
        size=Quantity.from_str(str(msg[1])),
        aggressor_side=AggressorSide.SELL if msg[2] == -1 else AggressorSide.BUY,
        trade_id=TradeId(str(msg[3])),  # zb future does not provide match id
        ts_event=secs_to_nanos(float(msg[3])),
        ts_init=ts_init,
    )


def parse_spot_book_snapshot_ws(
    instrument_id: InstrumentId, msg: Dict, ts_init: int
) -> OrderBookSnapshot:
    return OrderBookSnapshot(
        instrument_id=instrument_id,
        book_type=BookType.L2_MBP,
        bids=[[float(o[0]), float(o[1])] for o in msg.get("bids", [])],
        asks=[[float(o[0]), float(o[1])] for o in msg.get("asks", [])],
        # ts_event=ts_event,
        ts_event=secs_to_nanos(float(msg["timestamp"])),  # zb time precision too low
        ts_init=ts_init,
        update_id=msg["timestamp"],
    )


def parse_spot_trade_tick_ws(
    instrument_id: InstrumentId, msg: Dict[str, str], ts_init: int
) -> TradeTick:
    return TradeTick(
        instrument_id=instrument_id,
        price=Price.from_str(msg["price"]),
        size=Quantity.from_str(msg["amount"]),
        aggressor_side=AggressorSide.SELL if msg["type"] == "sell" else AggressorSide.BUY,
        trade_id=TradeId(str(msg["tid"])),
        # ts_event=secs_to_nanos(float(msg[3])),  # zb time precision too low
        ts_event=ts_init,
        ts_init=ts_init,
    )


def parse_trade_tick_ws(instrument_id: InstrumentId, msg: List, ts_init: int) -> TradeTick:
    return TradeTick(
        instrument_id=instrument_id,
        price=Price.from_str(str(msg[0])),
        size=Quantity.from_str(str(msg[1])),
        aggressor_side=AggressorSide.SELL if msg[2] == -1 else AggressorSide.BUY,
        trade_id=TradeId(str(msg[3])),
        # ts_event=secs_to_nanos(float(msg[3])),  # zb time not accurate
        ts_event=ts_init,
        ts_init=ts_init,
    )


def parse_bar(bar_type: BarType, values: List, ts_init: int) -> ZbBar:
    return ZbBar(
        bar_type=bar_type,
        open=Price.from_str(str(values[0])),
        high=Price.from_str(str(values[1])),
        low=Price.from_str(str(values[2])),
        close=Price.from_str(str(values[3])),
        volume=Quantity.from_str(str(values[4])),
        ts_event=secs_to_nanos(float(values[5])),
        ts_init=ts_init,
    )


def parse_spot_bar(bar_type: BarType, values: List, ts_init: int) -> ZbBar:
    return ZbBar(
        bar_type=bar_type,
        open=Price.from_str(str(values[1])),
        high=Price.from_str(str(values[2])),
        low=Price.from_str(str(values[3])),
        close=Price.from_str(str(values[4])),
        volume=Quantity.from_str(str(values[5])),
        ts_event=secs_to_nanos(float(values[0])),
        ts_init=ts_init,
    )


def parse_bar_ws(
    instrument_id: InstrumentId,
    kline: List,
    chan_type: str,
    ts_init: int,
) -> ZbBar:

    _, _, interval = chan_type.partition("_")
    resolution = interval[-1]
    if resolution == "M":
        aggregation = BarAggregation.MINUTE
    elif resolution == "H":
        aggregation = BarAggregation.HOUR
    elif resolution == "D":
        aggregation = BarAggregation.DAY
    else:  # pragma: no cover (design-time error)
        raise RuntimeError(f"unsupported time aggregation resolution, was {resolution}")

    bar_spec = BarSpecification(
        step=int(interval[:-1]),
        aggregation=aggregation,
        price_type=PriceType.LAST,
    )

    bar_type = BarType(
        instrument_id=instrument_id,
        bar_spec=bar_spec,
        aggregation_source=AggregationSource.EXTERNAL,
    )

    return ZbBar(
        bar_type=bar_type,
        open=Price.from_str(str(kline[0])),
        high=Price.from_str(str(kline[1])),
        low=Price.from_str(str(kline[2])),
        close=Price.from_str(str(kline[3])),
        volume=Quantity.from_str(str(kline[4])),
        ts_event=secs_to_nanos(float(kline[5])),
        ts_init=ts_init,
    )


# Order Parsing


def parse_spot_order_status(
    account_id: AccountId,
    instrument: Instrument,
    data: Dict[str, Any],
    report_id: UUID4,
    ts_init: int,
) -> OrderStatusReport:
    client_id_str = data.get("clientId")  # Not returned by zb spot
    price = data.get("price")

    created_at = millis_to_nanos(data["trade_date"])

    order_status = OrderStatus.ACCEPTED
    if data["status"] == 1:
        order_status = OrderStatus.CANCELED
    elif data["status"] == 2:
        order_status = (
            OrderStatus.CANCELED
            if data["trade_amount"] < data["total_amount"]
            else OrderStatus.FILLED
        )
    elif data["status"] == 3:
        order_status = OrderStatus.PARTIALLY_FILLED

    time_in_force = TimeInForce.GTC
    if data["type"] >= 4:
        time_in_force = TimeInForce.IOC

    return OrderStatusReport(
        account_id=account_id,
        instrument_id=instrument.id,
        client_order_id=ClientOrderId(client_id_str) if client_id_str is not None else None,
        venue_order_id=VenueOrderId(str(data["id"])),
        order_side=OrderSide.BUY if data["type"] % 2 == 1 else OrderSide.SELL,
        order_type=OrderType.LIMIT,  # ZB Spot only have limit order
        time_in_force=time_in_force,
        order_status=order_status,
        price=instrument.make_price(price) if price is not None else None,
        quantity=instrument.make_qty(data["total_amount"]),
        filled_qty=instrument.make_qty(data["trade_amount"]),
        avg_px=None,
        post_only=True if data["type"] == 2 or data["type"] == 3 else False,
        reduce_only=False,
        report_id=report_id,
        ts_accepted=created_at,
        ts_last=created_at,
        ts_init=ts_init,
    )


def parse_future_order_status(
    account_id: AccountId,
    instrument: Instrument,
    data: Dict[str, Any],
    report_id: UUID4,
    ts_init: int,
) -> OrderStatusReport:
    client_id_str = data.get("orderCode")
    price = data.get("price")
    avg_px = data["avgPrice"]
    created_at = millis_to_nanos(data["createTime"])

    order_status = OrderStatus.ACCEPTED
    if data["showStatus"] == 2:
        order_status = OrderStatus.PARTIALLY_FILLED
    elif data["showStatus"] == 3:
        order_status = OrderStatus.FILLED
    elif data["showStatus"] == 4:
        order_status = OrderStatus.PENDING_CANCEL
    elif data["showStatus"] in (5, 7):
        order_status = OrderStatus.CANCELED

    time_in_force = TimeInForce.GTC
    if data["action"] in (3, 31, 32, 33, 34, 39):
        time_in_force = TimeInForce.IOC

    return OrderStatusReport(
        account_id=account_id,
        instrument_id=instrument.id,
        client_order_id=ClientOrderId(client_id_str) if client_id_str is not None else None,
        venue_order_id=VenueOrderId(data["id"]),
        order_side=OrderSide.BUY if data["type"] == 1 else OrderSide.SELL,
        order_type=OrderType.LIMIT,  # ZB future only support limit order for now
        time_in_force=time_in_force,
        order_status=order_status,
        price=instrument.make_price(price) if price is not None else None,
        quantity=instrument.make_qty(data["amount"]),
        filled_qty=instrument.make_qty(data["tradeAmount"]),
        avg_px=Decimal(avg_px) if avg_px != "0" else None,
        post_only=True if data["action"] == 4 else False,
        reduce_only=False,  # Zb future not support reduce only
        report_id=report_id,
        ts_accepted=created_at,
        ts_last=created_at,
        ts_init=ts_init,
    )


def parse_order_fill(
    account_id: AccountId,
    instrument: Instrument,
    data: Dict[str, Any],
    report_id: UUID4,
    ts_init: int,
) -> TradeReport:
    return TradeReport(
        account_id=account_id,
        instrument_id=instrument.id,
        venue_order_id=VenueOrderId(str(data["orderId"])),
        trade_id=TradeId(str(data["createTime"])),
        order_side=OrderSide.BUY if data["side"] in (1, 4) else OrderSide.SELL,
        last_qty=instrument.make_qty(data["amount"]),
        last_px=instrument.make_price(data["price"]),
        commission=Money(data["feeAmount"], Currency.from_str(data["feeCurrency"])),
        liquidity_side=LiquiditySide.MAKER if data["maker"] else LiquiditySide.TAKER,
        report_id=report_id,
        ts_event=millis_to_nanos(data["createTime"]),
        ts_init=ts_init,
    )


def parse_position(
    account_id: AccountId,
    instrument: Instrument,
    data: Dict[str, Any],
    report_id: UUID4,
    ts_init: int,
) -> PositionStatusReport:
    if data["side"] == 2:
        position_side = PositionSide.LONG if data["nominalValue"] > 0 else PositionSide.SHORT
    elif data["side"] == 1:
        position_side = PositionSide.LONG
    else:  # data["side"] == 0:
        position_side = PositionSide.SHORT

    return PositionStatusReport(
        account_id=account_id,
        instrument_id=instrument.id,
        position_side=position_side,
        venue_position_id=PositionId(data["id"]),  # FIX: not same with self-generated position_id
        quantity=instrument.make_qty(data["amount"]),
        report_id=report_id,
        ts_last=ts_init,
        ts_init=ts_init,
    )


# Account parsing


def parse_zb_spot_liquidity_side(side: int) -> OrderSide:
    if side >= 4:  # IOC
        return LiquiditySide.TAKER
    elif side >= 2:  # Post only
        return LiquiditySide.MAKER
    else:  # NOTE: Limit Order might be taker sometime
        return LiquiditySide.MAKER


def parse_zb_spot_order_side(side: int) -> OrderSide:
    if side == 1 or side == 3 or side == 5:
        return OrderSide.BUY
    elif side == 0 or side == 2 or side == 4:
        return OrderSide.SELL
    else:
        raise ValueError(f"Unknown order side: {side}")


def parse_account_balances_raw(provider, raw_balances: List) -> List[AccountBalance]:
    return _parse_balances(provider, raw_balances, "currencyName", "amount", "freezeAmount")


def parse_spot_account_balances_raw(provider, raw_balances: List) -> List[AccountBalance]:
    return _parse_balances(provider, raw_balances, "enName", "available", "freez")


def _parse_balances(
    provider: InstrumentProvider,
    raw_balances: List[Dict[str, str]],
    asset_key: str,
    free_key: str,
    locked_key: str,
) -> List[AccountBalance]:
    parsed_balances: Dict[Currency, Tuple[Decimal, Decimal, Decimal]] = {}
    for b in raw_balances:
        currency = provider.currency(b[asset_key].upper())
        if not currency:
            continue

        free = Decimal(b[free_key])
        locked = Decimal(b[locked_key])
        total: Decimal = free + locked

        # Ignore empty balances
        if total == 0:
            continue

        parsed_balances[currency] = (total, locked, free)

    balances: List[AccountBalance] = [
        AccountBalance(
            total=Money(values[0], currency),
            locked=Money(values[1], currency),
            free=Money(values[2], currency),
        )
        for currency, values in parsed_balances.items()
    ]

    return balances


def parse_account_margins_raw(provider, raw_margins: List[Dict]) -> List[MarginBalance]:
    balances = []
    for margin in raw_margins:
        currency = provider.currency(margin["unit"].upper())
        balances.append(
            MarginBalance(
                initial=Money(margin["allMargin"], currency),
                maintenance=Money(margin["freeze"], currency),
            )
        )
    return balances


# ZB future use USDT as default


def parse_zb_position_side(side: int) -> PositionSide:
    if side == 3 or side == 1:
        return "LONG"
    else:
        return "SHORT"


def parse_zb_order_side(type: int) -> OrderSide:
    if type == 1:
        return OrderSide.BUY
    elif type == -1:
        return OrderSide.SELL
    else:
        raise ValueError(f"Unknown order side: {type}")


def parse_zb_order_type(action: int) -> OrderType:
    if action == 11 or action == 31 or action == 51:
        return OrderType.MARKET
    else:  # TODO: parse other STOP LIMIT
        return OrderType.LIMIT


def zb_order_side(  # noqa: C901
    order: Order, oms_type: OMSType, position: Optional[Position] = None
) -> int:
    if order.side == OrderSide.BUY:
        if oms_type == OMSType.NETTING:
            if order.is_reduce_only:
                return 0
            else:
                return 5
        elif oms_type == OMSType.HEDGING:
            # Closing short position
            if position is not None and position.side == PositionSide.SHORT:
                return 4
            else:
                return 1
    elif order.side == OrderSide.SELL:
        if oms_type == OMSType.NETTING:
            if order.is_reduce_only:
                return 0
            else:
                return 6
        elif oms_type == OMSType.HEDGING:
            # Closing long position
            if position is not None and position.side == PositionSide.LONG:
                return 3
            else:
                return 2
    raise ValueError(f"Unsupported oms type {oms_type}")


def zb_order_params(
    order: Order, oms_type: OMSType, position: Optional[Position] = None
) -> Dict[str, Any]:
    if order.type == OrderType.MARKET:
        return dict(
            symbol=order.instrument_id.symbol.value,
            side=zb_order_side(order, oms_type, position),
            amount=order.quantity.as_double(),
            action=zb_order_action_market(order),
            client_order_id=order.client_order_id.value,
        )
    elif order.type == OrderType.LIMIT:
        return dict(
            symbol=order.instrument_id.symbol.value,
            side=zb_order_side(order, oms_type, position),
            amount=order.quantity.as_double(),
            action=zb_order_action_limit(order),
            price=order.price.as_double(),
            client_order_id=order.client_order_id.value,
        )
    else:
        raise ValueError(f"Unsupported order type for zb: {order.type}")


def zb_order_action_market(order: MarketOrder) -> int:
    # Zb future using limit order to simulate market order
    if order.time_in_force == TimeInForce.GTC:
        return 19
    elif order.time_in_force == TimeInForce.IOC:
        return 39
    elif order.time_in_force == TimeInForce.FOK:
        return 59
    else:
        raise ValueError(f"Unsupported time_in_force market for zb: {order.time_in_force}")


def zb_order_action_limit(order: LimitOrder) -> int:
    if order.is_post_only:
        return 4
    elif order.time_in_force == TimeInForce.GTC:
        return 1
    elif order.time_in_force == TimeInForce.IOC:
        return 3
    elif order.time_in_force == TimeInForce.FOK:
        return 5
    else:
        raise ValueError(f"Unsupported time_in_force limit for zb: {order.time_in_force}")


def _quote_precision(quote_asset: str) -> int:
    #  common quotes are ("ZUSD", "USDT", "QC", "BTC", "USDC"):
    return 8  # hardcode for now


def parse_future_instrument_http(market: Dict, venue: Venue) -> Instrument:
    # Create base asset
    base_asset = market["sellerCurrencyName"].upper()
    base_currency = Currency(
        code=base_asset,
        precision=int(market["priceDecimal"]),
        iso4217=0,  # Currently undetermined for crypto assets
        name=base_asset,
        currency_type=CurrencyType.CRYPTO,
    )

    # Create quote asset
    quote_asset = market["buyerCurrencyName"].upper()
    quote_currency = Currency(
        code=quote_asset,
        precision=_quote_precision(quote_asset),
        iso4217=0,  # Currently undetermined for crypto assets
        name=quote_asset,
        currency_type=CurrencyType.CRYPTO,
    )

    symbol = Symbol(base_currency.code + "/" + quote_currency.code)
    # symbol = Symbol(market["symbol"])
    instrument_id = InstrumentId(symbol=symbol, venue=venue)

    # NOTE: ZB api doesn't provide info about tick size, simulate using price_precision
    price_increment = Price(
        value=1 / (10 ** int(market["priceDecimal"])), precision=int(market["priceDecimal"])
    )
    size_increment = Quantity(
        value=1 / (10 ** int(market["amountDecimal"])),
        precision=int(market["amountDecimal"]),
    )

    # TODO: query vip status programmatically
    # default fee rate is 0.075% for taker, 0.025% for vip taker
    # default fee rate is 0 for maker, -0.025% for vip maker
    maker_fee: Decimal = Decimal("-0.00025")  # vip
    taker_fee: Decimal = Decimal("0.00025")  # vip
    # maker_fee: Decimal = Decimal(0)
    # taker_fee: Decimal = Decimal("0.00075")

    return CryptoPerpetual(
        instrument_id=instrument_id,
        native_symbol=Symbol(market["symbol"]),
        base_currency=base_currency,
        quote_currency=quote_currency,
        settlement_currency=quote_currency,  # for USDT-based markets, settlement == quote
        is_inverse=False,
        price_precision=int(market["priceDecimal"]),
        size_precision=int(market["amountDecimal"]),
        price_increment=price_increment,
        size_increment=size_increment,
        max_quantity=Quantity(market["maxAmount"], int(market["amountDecimal"])),
        min_quantity=Quantity(market["minAmount"], int(market["amountDecimal"])),
        max_notional=None,
        min_notional=Money(market["minTradeMoney"], currency=quote_currency),
        max_price=None,
        min_price=None,
        margin_init=Decimal(0),
        margin_maint=Decimal(0),
        maker_fee=maker_fee,
        taker_fee=taker_fee,
        ts_event=time.time_ns(),
        ts_init=time.time_ns(),
        info=market,
    )


def parse_spot_instrument_http(market: str, entry: Dict, venue: Venue):
    base, _, quote = market.partition("_")

    # Create base asset
    base_asset = base.upper()
    base_currency = Currency(
        code=base_asset,
        precision=4,  # Hardcoded for zb spot
        iso4217=0,  # Currently undetermined for crypto assets
        name=base_asset,
        currency_type=CurrencyType.CRYPTO,
    )

    # Create quote asset
    quote_asset = quote.upper()
    quote_currency = Currency(
        code=quote_asset,
        precision=_quote_precision(quote_asset),
        iso4217=0,  # Currently undetermined for crypto assets
        name=quote_asset,
        currency_type=CurrencyType.CRYPTO,
    )

    symbol = Symbol(base_currency.code + "/" + quote_currency.code)
    instrument_id = InstrumentId(symbol=symbol, venue=venue)

    price_increment = Price(
        value=1 / (10 ** int(entry["priceScale"])),
        precision=int(entry["priceScale"]),
    )
    size_increment = Quantity(
        value=1 / (10 ** int(entry["amountScale"])),
        precision=int(entry["amountScale"]),
    )

    return CurrencyPair(
        instrument_id=instrument_id,
        native_symbol=Symbol(market),
        base_currency=base_currency,
        quote_currency=quote_currency,
        price_precision=int(entry["priceScale"]),
        size_precision=int(entry["amountScale"]),
        price_increment=price_increment,
        size_increment=size_increment,
        lot_size=None,
        max_quantity=None,
        min_quantity=Quantity(entry["minAmount"], int(entry["minAmount"])),
        max_notional=None,
        min_notional=None,
        max_price=None,
        min_price=None,
        margin_init=Decimal(0),
        margin_maint=Decimal(0),
        maker_fee=Decimal("0.002"),
        taker_fee=Decimal("0.002"),
        ts_event=time.time_ns(),
        ts_init=time.time_ns(),
        info=entry,
    )



# -------------------------------------------------------------------------------------------------
#  Copyright (C) 2015-2021 Nautech Systems Pty Ltd. All rights reserved.
#  https://nautechsystems.io
#
#  Licensed under the GNU Lesser General Public License Version 3.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at https://www.gnu.org/licenses/lgpl-3.0.en.html
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# -------------------------------------------------------------------------------------------------

from libc.stdint cimport int64_t
from nautilus_trader.common.providers cimport InstrumentProvider
from nautilus_trader.live.execution_client cimport LiveExecutionClient as NautilusLiveExecutionClient
from nautilus_trader.model.identifiers cimport AccountId
from nautilus_trader.model.identifiers cimport Venue


cdef class LiveExecutionClient(NautilusLiveExecutionClient):
    cdef readonly bint reconcile_by_fake_order

    cpdef void _set_venue(self, Venue venue) except *
    cpdef void _set_account_id(self, AccountId account_id) except *



from libc.stdint cimport int64_t
from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.model.data.tick cimport Tick
from nautilus_trader.model.identifiers cimport InstrumentId
from nautilus_trader.model.objects cimport Price


cdef class MarkTick(Tick):
    """
    Represents an instrument mark price at a venue.
    """

    def __init__(
        self,
        InstrumentId instrument_id not None,
        Price price not None,
        int64_t ts_event,
        int64_t ts_init,
    ):
        """
        Initialize a new instance of the ``MarkTick`` class.

        Parameters
        ----------
        price : Price
            The mark price for the instrument.
        ts_event : int64
            The UNIX timestamp (nanoseconds) when the close price event occurred.
        ts_init : int64
            The UNIX timestamp (nanoseconds) when the object was initialized.

        """
        super().__init__(instrument_id, ts_event, ts_init)
        self.price = price

    def __eq__(self, MarkTick other) -> bool:
        return MarkTick.to_dict_c(self) == MarkTick.to_dict_c(other)

    def __hash__(self) -> int:
        return hash(frozenset(MarkTick.to_dict_c(self)))

    def __str__(self) -> str:
        return (f"{self.instrument_id},"
                f"{self.price},"
                f"{self.ts_event}")

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self})"

    @staticmethod
    cdef MarkTick from_dict_c(dict values):
        Condition.not_none(values, "values")
        return MarkTick(
            instrument_id=InstrumentId.from_str_c(values["instrument_id"]),
            price=Price.from_str_c(values["price"]),
            ts_event=values["ts_event"],
            ts_init=values["ts_init"],
        )

    @staticmethod
    cdef dict to_dict_c(MarkTick obj):
        Condition.not_none(obj, "obj")
        return {
            "type": "MarkTick",
            "instrument_id": obj.instrument_id.value,
            "price": str(obj.price),
            "ts_event": obj.ts_event,
            "ts_init": obj.ts_init,
        }

    @staticmethod
    def from_dict(dict values) -> MarkTick:
        """
        Return an instrument close price event from the given dict values.

        Parameters
        ----------
        values : dict[str, object]
            The values for initialization.

        Returns
        -------
        MarkTick

        """
        return MarkTick.from_dict_c(values)

    @staticmethod
    def to_dict(MarkTick obj):
        """
        Return a dictionary representation of this object.

        Returns
        -------
        dict[str, object]

        """
        return MarkTick.to_dict_c(obj)



# -------------------------------------------------------------------------------------------------
#  Copyright (C) 2015-2021 Nautech Systems Pty Ltd. All rights reserved.
#  https://nautechsystems.io
#
#  Licensed under the GNU Lesser General Public License Version 3.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at https://www.gnu.org/licenses/lgpl-3.0.en.html
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# -------------------------------------------------------------------------------------------------

import asyncio
import types
from asyncio import Task
from typing import Callable, List, Optional

import aiohttp
from aiohttp import WSMessage

from nautilus_trader.common.logging cimport LogColor
from nautilus_trader.common.logging cimport Logger
from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.network.websocket cimport WebSocketClient as NautilusWebSocketClient

from nautilus_trader.network.error import MaxRetriesExceeded

from nacre.metrics.metrics import WEBSOCKET_DISCONNECT_COUNTER
from nacre.metrics.metrics import WEBSOCKET_RECONNECT_BACKOFF
from nacre.metrics.metrics import WEBSOCKET_RECV_ERROR_COUNTER


# from nautilus_trader.network.websocket cimport WSMsgType



cpdef enum WSMsgType:
    # websocket spec types
    CONTINUATION = 0x0
    TEXT = 0x1
    BINARY = 0x2
    PING = 0x9
    PONG = 0xA
    CLOSE = 0x8
    # aiohttp specific types
    CLOSING = 0x100
    CLOSED = 0x101
    ERROR = 0x102


cdef class WebSocketClient(NautilusWebSocketClient):
    def __init__(
        self,
        loop not None: asyncio.AbstractEventLoop,
        Logger logger not None: Logger,
        handler not None: Callable[[bytes], None],
        int max_retry_connection=0,
        bytes pong_msg=None,
        bint log_send=False,
        bint log_recv=False,
    ):
        super().__init__(
            loop=loop,
            logger=logger,
            handler=handler,
            max_retry_connection=max_retry_connection,
            pong_msg=pong_msg,
            log_send=log_send,
            log_recv=log_recv,
        )
        self.retrying = False

    async def receive(self) -> Optional[bytes]:
        cdef WSMsgType msg_type
        try:
            msg: WSMessage = await self._ws.receive()
            msg_type = msg.type
            if msg_type == TEXT:
                return msg.data.encode()  # Current workaround to always return bytes
            elif msg_type == BINARY:
                return msg.data
            elif msg_type == ERROR:  # aiohttp specific
                if self._stopping is True:
                    return
                self._log.warning(f"[RECV] {msg}.")
                raise ConnectionAbortedError("websocket aiohttp error")
            elif msg_type == CLOSE:  # Received CLOSE from server
                if self._stopping is True:
                    return
                self._log.warning(f"[RECV] {msg}.")
                raise ConnectionAbortedError("websocket closed by server")
            elif msg_type == CLOSING or msg_type == CLOSED:  # aiohttp specific
                if self._stopping is True:
                    return
                self._log.warning(f"[RECV] {msg}.")
                raise ConnectionAbortedError("websocket aiohttp closing or closed")
            else:
                self._log.warning(
                    f"[RECV] unknown data type: {msg.type}, data: {msg.data}.",
                )
                self.unknown_message_count += 1
                if self.unknown_message_count > 20:
                    self.unknown_message_count = 0  # Reset counter
                    # This shouldn't be happening, trigger a reconnection
                    raise ConnectionAbortedError("Too many unknown messages")
                return b""
        except (asyncio.IncompleteReadError, ConnectionAbortedError, RuntimeError) as ex:
            try:
                self.retrying = True
                WEBSOCKET_DISCONNECT_COUNTER.labels(endpoint=self._ws_url).inc()
                self._log.warning(
                    f"{ex.__class__.__name__}: Reconnecting {self.connection_retry_count=}, "
                    f"{self.max_retry_connection=}",
                )
                if self.max_retry_connection == 0:
                    raise
                if self.connection_retry_count > self.max_retry_connection:
                    raise MaxRetriesExceeded(f"Max retries of {self.max_retry_connection} exceeded.")
                await self._reconnect_backoff()
                self.connection_retry_count += 1
                self._log.debug(
                    f"Attempting reconnect (attempt: {self.connection_retry_count}).",
                )
                await self.reconnect()
            finally:
                self.retrying = False

    async def _reconnect_backoff(self) -> None:
        if self.connection_retry_count == 0:
            return  # Immediately attempt first reconnect
        cdef double backoff = 1.5 ** self.connection_retry_count
        WEBSOCKET_RECONNECT_BACKOFF.labels(endpoint=self._ws_url).set(backoff)
        self._log.debug(
            f"Exponential backoff attempt "
            f"{self.connection_retry_count}, sleeping for {backoff}",
        )
        await asyncio.sleep(backoff)

    async def start(self) -> None:
        self._log.debug("Starting recv loop...")
        cdef bytes raw
        while not self._stopping:
            try:
                raw = await self.receive()
                if self._log_recv:
                    self._log.info(f"[RECV] {raw}.", LogColor.BLUE)
                if raw is None:
                    continue
                if self._pong_msg is not None and raw == self._pong_msg:
                    continue  # Filter pong message
                self._handler(raw)
                self.connection_retry_count = 0
            except MaxRetriesExceeded as ex:
                self._log.exception("Error on max retry", ex)
                await self._reset_retry_count()  # Always retry don't stop!
                self._log.debug("Start new round.")
            except Exception as ex:
                WEBSOCKET_RECV_ERROR_COUNTER.labels(endpoint=self._ws_url).inc()
                self._log.exception(f"Error on receive", ex)
                # Always retry!!
                # break
        self._log.debug("Stopped.")
        self._stopped = True

    async def _reset_retry_count(self):
        nap = 2
        self.connection_retry_count = 0
        self._log.warning(f"Reset count and sleep for {nap} seconds")
        await asyncio.sleep(nap)



from nacre.serialization.base cimport AsyncSerializer


cdef class AvroSerializer(AsyncSerializer):
    cdef object _async_serializer
    cdef object _client

    cpdef dict type_convert(self, object obj, dict obj_dict)
    cpdef dict order_type_convert(self, object obj, dict obj_dict)
    cpdef dict position_type_convert(self, object obj, dict obj_dict)



import asyncio
import os
from functools import lru_cache
from typing import Dict, Optional, Union

from nautilus_trader.cache.cache import Cache
from nautilus_trader.common.clock import LiveClock
from nautilus_trader.common.logging import LiveLogger
from nautilus_trader.common.logging import Logger
from nautilus_trader.config import InstrumentProviderConfig
from nautilus_trader.live.factories import LiveDataClientFactory
from nautilus_trader.live.factories import LiveExecClientFactory
from nautilus_trader.model.identifiers import AccountId
from nautilus_trader.model.identifiers import Venue
from nautilus_trader.msgbus.bus import MessageBus

from nacre.adapters.zb.common.enums import ZbAccountType
from nacre.adapters.zb.config import ZbDataClientConfig
from nacre.adapters.zb.config import ZbExecClientConfig
from nacre.adapters.zb.data import ZbFutureDataClient
from nacre.adapters.zb.data import ZbSpotDataClient
from nacre.adapters.zb.futures.execution import ZbFuturesExecutionClient
from nacre.adapters.zb.http.client import ZbHttpClient
from nacre.adapters.zb.http.spot import ZbSpotHttpClient
from nacre.adapters.zb.providers import ZbInstrumentProvider
from nacre.adapters.zb.spot.execution import ZbSpotExecutionClient


HTTP_CLIENTS: Dict[str, ZbHttpClient] = {}


def get_cached_zb_http_client(
    loop: asyncio.AbstractEventLoop,
    clock: LiveClock,
    logger: Logger,
    base_url: str,
    key: Optional[str] = None,
    sec: Optional[str] = None,
    account_type: ZbAccountType = ZbAccountType.SPOT,
    proxy: Optional[str] = None,
) -> ZbHttpClient:

    global HTTP_CLIENTS
    key = key or os.environ.get("ZB_API_KEY", "")
    sec = sec or os.environ.get("ZB_API_SECRET", "")

    client_key: str = "|".join((key, sec, account_type.value, str(proxy or "")))
    if client_key not in HTTP_CLIENTS:
        if account_type.is_futures:
            client = ZbHttpClient(
                loop=loop,
                clock=clock,
                logger=logger,
                key=key,
                secret=sec,
                base_url=base_url,
                timeout=10,
                proxy=proxy,
            )
        elif account_type.is_spot:
            client = ZbSpotHttpClient(
                loop=loop,
                clock=clock,
                logger=logger,
                key=key,
                secret=sec,
                base_url=base_url,
                timeout=10,
                proxy=proxy,
            )
        else:
            raise ValueError(f"{account_type} not support by zb for now")

        HTTP_CLIENTS[client_key] = client
    return HTTP_CLIENTS[client_key]


@lru_cache(1)
def get_cached_zb_instrument_provider(
    logger: Logger,
    venue: Venue,
    client: ZbHttpClient,
    account_type: ZbAccountType,
    config: InstrumentProviderConfig,
) -> ZbInstrumentProvider:
    return ZbInstrumentProvider(
        logger=logger,
        venue=venue,
        client=client,
        account_type=account_type,
        config=config,
    )


class ZbLiveDataClientFactory(LiveDataClientFactory):
    @staticmethod
    def create(
        loop: asyncio.AbstractEventLoop,
        name: str,
        config: ZbDataClientConfig,
        msgbus: MessageBus,
        cache: Cache,
        clock: LiveClock,
        logger: LiveLogger,
    ) -> Union[ZbSpotDataClient, ZbFutureDataClient]:

        base_url_http_default: str = _get_default_http_base_url(config)
        base_url_ws_default: str = _get_default_ws_base_url(config)

        client = get_cached_zb_http_client(
            loop=loop,
            clock=clock,
            logger=logger,
            base_url=config.base_url_http or base_url_http_default,
            key=config.api_key,
            sec=config.api_secret,
            account_type=config.account_type,
            proxy=config.http_proxy,
        )

        venue = Venue(name.upper())

        # Get instrument provider singleton
        provider = get_cached_zb_instrument_provider(
            venue=venue,
            client=client,
            logger=logger,
            account_type=config.account_type,
            config=config.instrument_provider,
        )

        # Create client
        if config.account_type.is_spot:
            return ZbSpotDataClient(
                loop=loop,
                client=client,
                msgbus=msgbus,
                cache=cache,
                clock=clock,
                logger=logger,
                venue=venue,
                instrument_provider=provider,
                base_url_ws=config.base_url_ws or base_url_ws_default,
            )
        elif config.account_type.is_futures:
            return ZbFutureDataClient(
                loop=loop,
                client=client,
                msgbus=msgbus,
                cache=cache,
                clock=clock,
                logger=logger,
                venue=venue,
                instrument_provider=provider,
                base_url_ws=config.base_url_ws or base_url_ws_default,
            )
        else:
            raise ValueError(f"Account type not implemented: {config.account_type}")


class ZbLiveExecClientFactory(LiveExecClientFactory):
    @staticmethod
    def create(
        loop: asyncio.AbstractEventLoop,
        name: str,
        config: ZbExecClientConfig,
        msgbus: MessageBus,
        cache: Cache,
        clock: LiveClock,
        logger: LiveLogger,
    ) -> Union[ZbSpotExecutionClient, ZbFuturesExecutionClient]:

        base_url_http_default: str = _get_default_http_base_url(config)
        base_url_ws_default: str = _get_default_ws_base_url(config)

        client = get_cached_zb_http_client(
            loop=loop,
            clock=clock,
            logger=logger,
            base_url=config.base_url_http or base_url_http_default,
            key=config.api_key,
            sec=config.api_secret,
            account_type=config.account_type,
            proxy=config.http_proxy,
        )

        res = name.partition("-")
        if res[1]:
            account_id = AccountId.from_str(name)
        else:
            account_id = AccountId(name, "DEFAULT")
        venue = Venue(account_id.issuer)

        market_client = client
        # ZbSpot separate market and trade api to "api.zb.com" and "trade.zb.com"
        if config.account_type.is_spot:
            market_client = get_cached_zb_http_client(
                loop=loop,
                clock=clock,
                logger=logger,
                base_url="https://api.zb.com",  # Hardcode for now
                account_type=config.account_type,
            )

        provider = get_cached_zb_instrument_provider(
            venue=venue,
            client=market_client,
            logger=logger,
            account_type=config.account_type,
            config=config.instrument_provider,
        )

        if config.account_type.is_spot:
            return ZbSpotExecutionClient(
                loop=loop,
                client=client,
                market_client=market_client,
                name=name,
                account_id=account_id,
                venue=venue,
                msgbus=msgbus,
                cache=cache,
                clock=clock,
                logger=logger,
                instrument_provider=provider,
                base_url_ws=config.base_url_ws or base_url_ws_default,
            )
        elif config.account_type.is_futures:
            return ZbFuturesExecutionClient(
                loop=loop,
                client=client,
                oms_type=config.oms_type,
                name=name,
                account_id=account_id,
                venue=venue,
                msgbus=msgbus,
                cache=cache,
                clock=clock,
                logger=logger,
                instrument_provider=provider,
                base_url_ws=config.base_url_ws or base_url_ws_default,
            )
        else:
            raise ValueError(f"Account type not implemented: {config.account_type}")


def _get_default_http_base_url(config: Union[ZbDataClientConfig, ZbExecClientConfig]) -> str:
    if config.account_type == ZbAccountType.SPOT:
        if isinstance(config, ZbDataClientConfig):
            return "https://api.zb.com"
        elif isinstance(config, ZbExecClientConfig):
            return "https://trade.zb.com"
    elif config.account_type == ZbAccountType.FUTURES_USDT:
        return "https://fapi.zb.com"

    raise RuntimeError(f"invalid Zb account type, was {config.account_type}")


def _get_default_ws_base_url(config: Union[ZbDataClientConfig, ZbExecClientConfig]) -> str:
    if config.account_type in (ZbAccountType.SPOT, ZbAccountType.MARGIN):
        return "wss://api.zb.com/websocket"
    elif config.account_type == ZbAccountType.FUTURES_USDT:
        if isinstance(config, ZbDataClientConfig):
            return "wss://fapi.zb.com/ws/public/v1"
        elif isinstance(config, ZbExecClientConfig):
            return "wss://fapi.zb.com/ws/private/api/v2"

    raise RuntimeError(f"invalid Zb account type, was {config.account_type}")



import hashlib
import hmac
import time
from typing import Any, Dict, List, Optional, Union

import nautilus_trader
from aiohttp import ClientConnectorError

from nacre.adapters.zb.http.client import ZbHttpClient
from nacre.adapters.zb.http.error import ZbSpotOperationError


NAUTILUS_VERSION = nautilus_trader.__version__


def retry_if_connect_error(exception):
    return isinstance(exception, ClientConnectorError)


class ZbSpotHttpClient(ZbHttpClient):

    BASE_URL = "https://api.zb.com"  # market api
    # BASE_URL = "https://trade.zb.com"  # trade api

    async def sign_request(
        self,
        http_method: str,
        url_path: str,
        payload: Optional[Dict[str, Any]] = None,
    ) -> Any:
        if payload is None:
            payload = {}
        payload["accesskey"] = self._key
        query_string = self._prepare_params(payload)
        payload["sign"] = hmac.new(
            bytes(self._hashed_secret, encoding="utf-8"), query_string.encode("utf-8"), hashlib.md5
        ).hexdigest()
        payload["reqTime"] = (int)(time.time() * 1000)

        return await self.send_request(http_method, url_path, payload)

    def _raise_if_operation_code_error(self, resp: Union[Dict, List]):
        if not isinstance(resp, dict):
            return

        code = resp.get("code")
        if code is None:
            return
        elif code == 1000:
            return
        else:
            raise ZbSpotOperationError(status=code, message=resp.get("message", str(resp)))



import asyncio
import re
from datetime import datetime
from typing import Any, Dict, List, Optional

import orjson
import pandas as pd
from nautilus_trader.cache.cache import Cache
from nautilus_trader.common.clock import LiveClock
from nautilus_trader.common.logging import LogColor
from nautilus_trader.common.logging import Logger
from nautilus_trader.core.datetime import millis_to_nanos

# from nautilus_trader.model.commands.trading import CancelAllOrders
from nautilus_trader.execution.messages import CancelOrder
from nautilus_trader.execution.messages import SubmitOrder
from nautilus_trader.execution.reports import OrderStatusReport
from nautilus_trader.execution.reports import PositionStatusReport
from nautilus_trader.execution.reports import TradeReport
from nautilus_trader.model.enums import AccountType
from nautilus_trader.model.enums import OMSType
from nautilus_trader.model.enums import OrderSide
from nautilus_trader.model.enums import OrderType
from nautilus_trader.model.enums import TimeInForce
from nautilus_trader.model.identifiers import AccountId
from nautilus_trader.model.identifiers import ClientId
from nautilus_trader.model.identifiers import InstrumentId
from nautilus_trader.model.identifiers import TradeId
from nautilus_trader.model.identifiers import Venue
from nautilus_trader.model.identifiers import VenueOrderId
from nautilus_trader.model.instruments.base import Instrument
from nautilus_trader.model.objects import AccountBalance
from nautilus_trader.model.objects import MarginBalance
from nautilus_trader.model.objects import Money
from nautilus_trader.model.objects import Price
from nautilus_trader.model.objects import Quantity
from nautilus_trader.model.orders.base import Order
from nautilus_trader.model.orders.limit import LimitOrder
from nautilus_trader.msgbus.bus import MessageBus

from nacre.adapters.zb.common import format_websocket_user_market
from nacre.adapters.zb.http.api.spot_account import ZbSpotAccountHttpAPI
from nacre.adapters.zb.http.api.spot_market import ZbSpotMarketHttpAPI
from nacre.adapters.zb.http.client import ZbHttpClient
from nacre.adapters.zb.http.error import ZbError
from nacre.adapters.zb.parsing import parse_spot_account_balances_raw
from nacre.adapters.zb.parsing import parse_spot_order_status
from nacre.adapters.zb.parsing import parse_zb_spot_liquidity_side
from nacre.adapters.zb.parsing import parse_zb_spot_order_side
from nacre.adapters.zb.providers import ZbInstrumentProvider
from nacre.adapters.zb.websocket.spot import ZbSpotUserDataWebSocket
from nacre.live.execution_client import LiveExecutionClient


class ZbSpotExecutionClient(LiveExecutionClient):
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        client: ZbHttpClient,
        market_client: ZbHttpClient,
        name: str,
        account_id: AccountId,
        venue: Venue,
        msgbus: MessageBus,
        cache: Cache,
        clock: LiveClock,
        logger: Logger,
        instrument_provider: ZbInstrumentProvider,
        base_url_ws: Optional[str] = None,
    ):
        # HTTP API
        self._account_api = ZbSpotAccountHttpAPI(client=client)
        self._market_api = ZbSpotMarketHttpAPI(client=market_client)

        # WebSocket API
        self._ws_user_api = ZbSpotUserDataWebSocket(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=self._handle_user_ws_message,
            key=client.api_key,
            hashed_secret=client.hashed_secret,
            base_url=base_url_ws,
        )

        self._client = client
        self._proxy = client._proxy
        # Hot caches
        self._instrument_ids: Dict[str, InstrumentId] = {}
        self._order_types: Dict[VenueOrderId, OrderType] = {}

        super().__init__(
            loop=loop,
            client_id=ClientId(name),
            venue=venue,
            oms_type=OMSType.NETTING,
            instrument_provider=instrument_provider,
            account_type=AccountType.CASH,
            base_currency=None,
            msgbus=msgbus,
            cache=cache,
            clock=clock,
            logger=logger,
            config={"name": f"ZbExecClient-{name}"},
        )

        self._set_account_id(account_id)
        self._set_venue(Venue(account_id.issuer))

    def connect(self) -> None:
        self._log.info("Connecting...")
        self._loop.create_task(self._connect())

    def disconnect(self) -> None:
        self._log.info("Disconnecting...")
        self._loop.create_task(self._disconnect())

    async def _connect(self) -> None:
        if not self._client.connected:
            await self._client.connect()
        try:
            await self._instrument_provider.initialize()
            response = await self._account_api.get_account()
        except Exception as ex:
            self._log.exception("Error on connect", ex)
            return

        self._authenticate_api_key(response)
        balances = self.parse_account_balances(response)
        self._update_account_state(balances)

        # Connect WebSocket clients
        await self._connect_websockets()

        self._set_connected(True)
        assert self.is_connected
        self._log.info("Connected.")

    async def _connect_websockets(self) -> None:
        kwargs = {"proxy": self._proxy} if self._proxy else {}
        await self._ws_user_api.connect(**kwargs)

        await self._ws_user_api.subscribe_asset_snapshot()
        await self._subscribe_order_update_on_selected_markets()
        # await self._ws_user_api.logged_in()

    async def _disconnect(self) -> None:
        # Disconnect WebSocket clients
        if self._ws_user_api.is_connected:
            await self._ws_user_api.disconnect()

        if self._client.connected:
            await self._client.disconnect()

        self._set_connected(False)
        self._log.info("Disconnected.")

    def _update_account_state(
        self, balances: List[AccountBalance], margins: Optional[List[MarginBalance]] = None
    ) -> None:
        if not balances:
            return

        self.generate_account_state(
            balances=balances,
            margins=[] if margins is None else margins,
            reported=True,
            ts_event=self._clock.timestamp_ns(),  # zb account doesn't provide updateTime
        )

    async def _verify_order_status(self, order: Order):
        try:
            response = await self._account_api.get_order(
                symbol=order.instrument_id.symbol.value,
                client_order_id=order.client_order_id,
            )
            self._log.debug(f"Fetched order status: {response}")

            venue_order_id = VenueOrderId(response["id"])
            self.generate_order_accepted(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                venue_order_id=venue_order_id,
                ts_event=self._clock.timestamp_ns(),
            )
        except ZbError as ex:
            self.generate_order_rejected(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                reason=ex.message,
                ts_event=self._clock.timestamp_ns(),
            )

    # -- EXECUTION REPORTS -------------------------------------------------------------------------

    async def generate_order_status_report(
        self,
        venue_order_id: VenueOrderId = None,
    ) -> Optional[OrderStatusReport]:
        if venue_order_id is None:
            return None
        return None

        # ZB spot get order require symbol

        # try:
        #     response = await self._account_api.get_order(order_id=venue_order_id.value)
        # except ZbError as ex:
        #     self._log.error(ex.message)  # type: ignore  # TODO(cs): Improve errors
        #     return None
        #
        # # Get instrument
        # instrument_id: InstrumentId = self._get_cached_instrument_id(response)
        # instrument = self._instrument_provider.find(instrument_id)
        # if instrument is None:
        #     self._log.error(
        #         f"Cannot generate order status report: "
        #         f"no instrument found for {instrument_id}.",
        #     )
        #     return None
        #
        # return self.parse_order_status(
        #     account_id=self.account_id,
        #     instrument=instrument,
        #     data=response,
        #     report_id=self._uuid_factory.generate(),
        #     ts_init=self._clock.timestamp_ns(),
        # )

    async def generate_order_status_reports(
        self,
        instrument_id: InstrumentId = None,
        start: datetime = None,
        end: datetime = None,
        open_only: bool = False,
    ) -> List[OrderStatusReport]:
        self._log.info(f"Generating OrderStatusReports for {self.id}...")

        reports: List[OrderStatusReport] = []
        reports += await self._get_order_status_reports(
            instrument_id=instrument_id,
            start=start,
            end=end,
            open_only=open_only,
        )

        len_reports = len(reports)
        plural = "" if len_reports == 1 else "s"
        self._log.info(f"Generated {len(reports)} OrderStatusReport{plural}.")

        return reports

    async def generate_trade_reports(
        self,
        instrument_id: InstrumentId = None,
        venue_order_id: VenueOrderId = None,
        start: datetime = None,
        end: datetime = None,
    ) -> List[TradeReport]:
        self._log.info(f"Generating TradeReports for {self.id}...")

        # reports: List[TradeReport] = []

        reports = await self._get_trade_reports(
            instrument_id=instrument_id,
            venue_order_id=venue_order_id,
            start=start,
            end=end,
        )

        reports = sorted(reports, key=lambda x: x.trade_id)

        len_reports = len(reports)
        plural = "" if len_reports == 1 else "s"
        self._log.info(f"Generated {len(reports)} TradeReport{plural}.")

        return reports

    async def generate_position_status_reports(
        self,
        instrument_id: InstrumentId = None,
        start: datetime = None,
        end: datetime = None,
    ) -> List[PositionStatusReport]:

        self._log.info(f"Generating PositionStatusReports for {self.id}...")
        return []

    async def _subscribe_order_update_on_selected_markets(self):
        # ZB spot require subscribe each market specifically
        # TODO: better way to escape zb ws subscribe rate limit 60 sub/sec
        for market in self._instrument_provider.list_local_market_ids():
            if self._instrument_provider._filters:
                if not re.match(self._instrument_provider._filters["subscribe"], market):
                    self._log.debug(f"Ignored order update on {market}")
                    continue
            await asyncio.sleep(1 / 60)
            await self._ws_user_api.subscribe_order_update(f"{market}default")

    def parse_account_balances(self, response: Dict) -> List[AccountBalance]:
        return parse_spot_account_balances_raw(
            self._instrument_provider, response["result"]["coins"]
        )

    def _get_cached_instrument_id(self, response: Dict) -> InstrumentId:
        return self._instrument_provider.find_instrument_id_by_local_market_id(response["currency"])

    def _authenticate_api_key(self, response: Dict[str, Any]) -> None:
        if response["entrustPerm"]:
            self._log.info("Zb API key authenticated.", LogColor.GREEN)
            self._log.info(f"API key {self._client.api_key} has trading permissions.")
        else:
            self._log.error("Zb API key does not have trading permissions.")

    def _handle_user_ws_message(self, raw: bytes):
        try:
            msg: Dict[str, Any] = orjson.loads(raw)
        except orjson.JSONDecodeError:
            plain_text = raw.decode()
            if plain_text == "pong":
                return
            else:
                self._log.error(f"Unknown ws message received: {plain_text}")
                return

        chan: Optional[str] = msg.get("channel")
        code = msg.get("code")
        if code is not None and code != 1000:
            self._log.error(f"Error subscribe: {msg}")
            return

        if chan == "push_user_incr_record":
            self._handle_order_update(msg)
        elif chan == "push_user_asset":
            self._handle_asset_update(msg)
        else:
            self._log.warning(f"Unrecognized websocket message type, msg {msg}")
            return

    def _handle_order_update(self, data: Dict[str, Any]):
        self._log.debug(f"Received order update {data}")
        record: List = data["record"]

        # Parse client order ID
        venue_order_id = VenueOrderId(record[0])
        client_order_id = self._cache.client_order_id(venue_order_id)
        if client_order_id is None:
            self._log.warning(f"Cannot fill un-cached order with {repr(venue_order_id)}")
            return

        # Fetch strategy ID
        strategy_id = self._cache.strategy_id_for_order(client_order_id)
        if strategy_id is None:
            self._log.error(
                f"Cannot handle order update: " f"strategy ID for {client_order_id} not found",
            )
            return

        ts_event: int = millis_to_nanos(record[16])

        # Parse instrument ID
        zb_symbol = format_websocket_user_market(data["market"])
        instrument_id = self._instrument_provider.find_instrument_id_by_local_market_id(zb_symbol)

        order_status = record[7]
        if order_status == 1:  # 1 for cancel
            self.generate_order_canceled(
                strategy_id=strategy_id,
                instrument_id=instrument_id,
                client_order_id=client_order_id,
                venue_order_id=venue_order_id,
                ts_event=ts_event,
            )
        elif order_status == 2 or (
            order_status == 3 and record[3] > 0
        ):  # 2 for filled/partial filled complete
            instrument: Instrument = self._instrument_provider.find(instrument_id)

            self.generate_order_filled(
                strategy_id=strategy_id,
                instrument_id=instrument_id,
                client_order_id=client_order_id,
                venue_order_id=venue_order_id,
                venue_position_id=None,
                trade_id=TradeId(str(record[13])),
                order_side=parse_zb_spot_order_side(record[5]),
                order_type=OrderType.LIMIT,
                last_qty=Quantity.from_str(str(record[15])),
                last_px=Price.from_str(str(record[14])),
                quote_currency=instrument.quote_currency,
                commission=Money(record[8], instrument.quote_currency),
                liquidity_side=parse_zb_spot_liquidity_side(record[5]),
                ts_event=ts_event,
            )
        elif order_status == 0 or order_status == 3:  # 0 for pending, 3 for partial fill
            self.generate_order_accepted(
                strategy_id=strategy_id,
                instrument_id=instrument_id,
                client_order_id=client_order_id,
                venue_order_id=venue_order_id,
                ts_event=ts_event,
            )

    def _handle_asset_update(self, data: Dict):
        balances = parse_spot_account_balances_raw(self._instrument_provider, data["coins"])
        self._update_account_state(balances)

    def submit_order(self, command: SubmitOrder) -> None:
        order: Order = command.order
        self._loop.create_task(self._submit_order(order))

    def cancel_order(self, command: CancelOrder) -> None:
        self._loop.create_task(self._cancel_order(command))

    # def cancel_all_orders(self, command: CancelAllOrders) -> None:
    #     self._loop.create_task(self._cancel_all_order(command))

    async def _submit_order(self, order: Order) -> None:
        self._log.debug(f"Submitting {order}.")

        if order.type != OrderType.LIMIT:
            self._log.error(
                "Cannot submit order: "
                f"{order.type} orders not supported by the exchange for SPOT markets. "
                "Only LIMIT supported."
            )
            return

        # Generate event here to ensure correct ordering of events
        self.generate_order_submitted(
            strategy_id=order.strategy_id,
            instrument_id=order.instrument_id,
            client_order_id=order.client_order_id,
            ts_event=self._clock.timestamp_ns(),
        )

        try:
            response = await self._submit_limit_order(order)
            self._log.debug(f"Submit order response {response}")
            venue_order_id = VenueOrderId(response["id"])
            self.generate_order_accepted(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                venue_order_id=venue_order_id,
                ts_event=self._clock.timestamp_ns(),
            )
        except ZbError as ex:
            self.generate_order_rejected(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                reason=ex.message,
                ts_event=self._clock.timestamp_ns(),
            )
        except Exception as ex:
            self._log.error(f"Error occurs in submit order: {ex}, verifying order ...")
            await self._verify_order_status(order)

    async def _submit_limit_order(self, order: LimitOrder) -> Dict:
        order_type = None
        if order.is_post_only:
            order_type = 1
        elif order.time_in_force == TimeInForce.IOC:
            order_type = 2

        return await self._account_api.order(
            amount=order.quantity.as_double(),
            currency=order.instrument_id.symbol.value,
            customer_order_id=order.client_order_id.value,
            price=order.price.as_double(),
            trade_type=1 if order.side == OrderSide.BUY else 0,
            order_type=order_type,
        )

    async def _cancel_order(self, order: CancelOrder) -> None:
        self._log.debug(f"Canceling order {order.client_order_id.value}.")

        self.generate_order_pending_cancel(
            strategy_id=order.strategy_id,
            instrument_id=order.instrument_id,
            client_order_id=order.client_order_id,
            venue_order_id=order.venue_order_id,
            ts_event=self._clock.timestamp_ns(),
        )

        try:
            response = await self._account_api.cancel_order(
                currency=order.instrument_id.symbol.value,
                id=order.venue_order_id.value,
            )
            self._log.debug(f"Cancel order response {response}")
        except Exception as ex:
            if isinstance(ex, ZbError):
                message = ex.message
            else:
                message = repr(ex)
                self._log.exception("Error on cancel order", ex)
            self.generate_order_cancel_rejected(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                venue_order_id=order.venue_order_id,
                reason=message,
                ts_event=self._clock.timestamp_ns(),  # TODO(cs): Parse from response
            )

    async def _get_order_status_reports(
        self,
        instrument_id: InstrumentId = None,
        start: datetime = None,
        end: datetime = None,
        open_only: bool = False,
    ) -> List[OrderStatusReport]:
        reports: List[OrderStatusReport] = []

        if instrument_id is None:
            self._log.warning("InstrumentId not specified, zb does not support all orders query")
            return reports

        try:
            if open_only:
                response: List[
                    Dict[str, Any]
                ] = await self._account_api.get_unfinished_orders_ignore_trade_type(
                    currency=instrument_id.symbol.value,
                )
            else:
                response = await self._account_api.get_orders_ignore_trade_type(
                    currency=instrument_id.symbol.value,
                )
        except ZbError as ex:
            self._log.error(ex.message)  # type: ignore  # TODO(cs): Improve errors
            return []

        if response:
            for data in response:
                created_at = pd.to_datetime(data["trade_date"], unit="ms")
                if start is not None and created_at < start:
                    continue
                if end is not None and created_at > end:
                    continue

                # Get instrument
                instrument_id = instrument_id or self._get_cached_instrument_id(data)
                instrument = self._instrument_provider.find(instrument_id)
                if instrument is None:
                    self._log.error(
                        f"Cannot generate order status report: "
                        f"no instrument found for {instrument_id}.",
                    )
                    continue

                report: OrderStatusReport = parse_spot_order_status(
                    account_id=self.account_id,
                    instrument=instrument,
                    data=data,
                    report_id=self._uuid_factory.generate(),
                    ts_init=self._clock.timestamp_ns(),
                )

                self._log.debug(f"Received {report}.")
                reports.append(report)

        return reports

    async def _get_trade_reports(
        self,
        instrument_id: InstrumentId = None,
        venue_order_id: VenueOrderId = None,
        start: datetime = None,
        end: datetime = None,
    ) -> List[TradeReport]:
        return []  # Zb spot does not provide trade detail

    async def _get_position_status_report(
        self,
        instrument_id: InstrumentId = None,
        start: datetime = None,
        end: datetime = None,
    ) -> List[PositionStatusReport]:
        return []



from cpython.datetime cimport datetime
from cpython.datetime cimport timedelta
from nautilus_trader.indicators.base.indicator cimport Indicator
from nautilus_trader.model.identifiers cimport InstrumentId
from nautilus_trader.model.objects cimport Price


cdef class BarMinMax(Indicator):
    cdef readonly timedelta lookback
    """The look back duration in time.\n\n:returns: `timedelta`"""
    cdef readonly WindowedMinMaxPrices prices
    """The windowed min max prices.\n\n:returns: `WindowedMinMaxPrices`"""
    cdef readonly WindowedMinMaxPrices volumes
    """The windowed min max prices.\n\n:returns: `WindowedMinMaxPrices`"""

cdef class WindowedMinMaxPrices:
    cdef object _min_prices
    cdef object _max_prices

    cdef readonly timedelta lookback
    """The look back duration in time.\n\n:returns: `timedelta`"""
    cdef readonly Price min_price
    """The minimum price in the window.\n\n:returns: `Price`"""
    cdef readonly Price max_price
    """The maximum price in the window.\n\n:returns: `Price`"""

    cpdef void add_price(self, datetime ts, Price price) except *
    cpdef void reset(self) except *

    cdef void _expire_stale_prices_by_cutoff(self, ts_prices, datetime cutoff) except *
    cdef void _add_min_price(self, datetime ts, Price price) except *
    cdef void _add_max_price(self, datetime ts, Price price) except *



from typing import Callable

from tenacity import RetryCallState
from tenacity import _utils


def after_log() -> Callable[["RetryCallState"], None]:
    """After call strategy that logs to some logger the finished attempt."""

    def log_it(retry_state: "RetryCallState") -> None:
        args = retry_state.args
        if len(args) > 0 and hasattr(args[0], "_log"):
            args[0]._log.warning(f"Error: {repr(retry_state.outcome.exception())} ")

            args[0]._log.warning(
                f"Finished call to '{_utils.get_callback_name(retry_state.fn)}' "
                f"after {retry_state.seconds_since_start}(s), "
                f"this was the {_utils.to_ordinal(retry_state.attempt_number)} time calling it.",
            )

    return log_it



#!/usr/bin/env python3
from nautilus_trader.config import InstrumentProviderConfig

# from nautilus_trader.config import CacheDatabaseConfig
from nautilus_trader.live.node import TradingNode
from nautilus_trader.model.enums import OMSType

# from nacre.config.actors import KafkaConfig
from nacre.adapters.zb.common.enums import ZbAccountType
from nacre.adapters.zb.config import ZbDataClientConfig
from nacre.adapters.zb.config import ZbExecClientConfig
from nacre.adapters.zb.factories import ZbLiveDataClientFactory
from nacre.adapters.zb.factories import ZbLiveExecClientFactory

# from nacre.infrastructure.kafka import KafkaPubSub
from nacre.config import TradingNodeConfig
from nacre.examples.strategies.mon import Mon
from nacre.examples.strategies.mon import MonConfig


# os.environ["PYTHONASYNCIODEBUG"] = "1"
# import logging
# logging.basicConfig(level=logging.DEBUG)

# sys.stdout = open("/var/log/nacre/watcher.log", 'w')
# sys.stderr = sys.stdout


# Configure the trading node
config_node = TradingNodeConfig(
    trader_id="MM-001",
    log_level="DEBUG",
    data_clients={
        "ZB": ZbDataClientConfig(
            account_type=ZbAccountType.FUTURES_USDT,
            # instrument_provider=InstrumentProviderConfig(load_ids=["ZETH/ZUSD.ZB"]),
            instrument_provider=InstrumentProviderConfig(load_all=True),
            base_url_http="https://futures.zb.zone",
            base_url_ws="wss://futures.zb.zone/ws/public/v1",
        ),
        # "ZB": ZbDataClientConfig(
        #     account_type=ZbAccountType.SPOT,
        #     # instrument_provider=InstrumentProviderConfig(load_ids=["ZETH/ZUSD.ZB"]),
        #     instrument_provider=InstrumentProviderConfig(load_all=True),
        # ),
    },
    exec_clients={
        "ZB": ZbExecClientConfig(  # noqa: S106
            api_key="",
            api_secret="",
            oms_type=OMSType.NETTING,
            account_type=ZbAccountType.FUTURES_USDT,
            instrument_provider=InstrumentProviderConfig(load_all=True),
            # base_url_http="https://futures.zb.zone",
            # base_url_ws="wss://futures.zb.zone/ws/private/api/v2",
        ),
    },
    # cache_database=CacheDatabaseConfig(),
    # timeout_disconnection=30,
    # timeout_connection=30,
)
# Instantiate the node with a configuration
node = TradingNode(config=config_node)


# actor = KafkaPubSub(
#     config=KafkaConfig(
#         bootstrap_servers="192.168.18.4:19092,192.168.18.4:29092,192.168.18.4:39092",
#         schema_registry="http://192.168.18.4:8081",
#         topic_filters=["events.order*", "events.position*"],
#     )
# )
# node.trader.add_actor(actor)

stra = Mon(
    config=MonConfig(
        venues="ZB",
        oms_type="NETTING",
    )
)
node.trader.add_strategy(stra)

node.add_data_client_factory("ZB", ZbLiveDataClientFactory)
node.add_exec_client_factory("ZB", ZbLiveExecClientFactory)
node.build()

# Stop and dispose of the node with SIGINT/CTRL+C
if __name__ == "__main__":
    try:
        node.start()
    finally:
        node.dispose()



import importlib
import os
import sys
from typing import Dict, Optional

import fire
import yaml  # type: ignore
from nautilus_trader.common.actor import Actor
from nautilus_trader.config import ActorFactory
from nautilus_trader.config import ImportableActorConfig
from nautilus_trader.config import ImportableStrategyConfig
from nautilus_trader.config import StrategyFactory
from nautilus_trader.live.node import TradingNode
from nautilus_trader.trading.strategy import TradingStrategy

from nacre.config import TradingNodeConfig


def run():
    fire.Fire(run_live_node)


def run_live_node(config: str, log_path: Optional[str] = None, credential_path: str = "~/.nsrc"):
    with open(config, "r") as stream:
        node_cfg = yaml.safe_load(stream)

    # redirect log
    if log_path is None and "log_path" in node_cfg:
        log_path = node_cfg.pop("log_path")
    if log_path is not None:
        sys.stdout = open(log_path, "w")
        sys.stderr = sys.stdout

    # insert path
    paths = node_cfg.pop("paths", None)
    if paths and isinstance(paths, list):
        for path in paths:
            sys.path.insert(0, path)

    # load api key/sec from default credential path
    _load_credentials(node_cfg, credential_path)

    factories_cfg = node_cfg.pop("factories")

    # initialize actors
    actors = [_build_actor(actor_config) for actor_config in node_cfg.pop("actors", [])]
    # initialize strategies
    strategies = [_build_strategy(strategy) for strategy in node_cfg.pop("strategies", [])]

    node = TradingNode(config=TradingNodeConfig.parse_obj(node_cfg))
    if actors:
        node.trader.add_actors(actors)
    if strategies:
        node.trader.add_strategies(strategies)

    # apply factories
    _add_client_factories(node, factories_cfg)

    node.build()
    try:
        node.start()
    finally:
        node.dispose()


def _import_cls(path: str):
    assert path
    assert ":" in path

    module, cls = path.rsplit(":")
    mod = importlib.import_module(module)
    return getattr(mod, cls)


def _add_client_factories(node: TradingNode, config: Dict):
    if "data_clients" in config:
        for venue, path in config["data_clients"].items():
            node.add_data_client_factory(venue, _import_cls(path))
    if "exec_clients" in config:
        for venue, path in config["exec_clients"].items():
            node.add_exec_client_factory(venue, _import_cls(path))


def _build_actor(actor: Dict) -> Actor:
    config = ImportableActorConfig(**actor)
    return ActorFactory.create(config)


def _build_strategy(strategy: Dict) -> TradingStrategy:
    config = ImportableStrategyConfig(**strategy)
    return StrategyFactory.create(config)


def _load_credentials(node_cfg: Dict, credential_path: str) -> None:
    """
    Fulfill exec_clients with exec_client_ids

    """
    exec_clients = node_cfg.get("exec_clients", {})
    exec_client_ids = node_cfg.pop("exec_client_ids", None)
    if not exec_client_ids:
        return

    with open(os.path.expanduser(credential_path)) as yamlfile:
        account_credentials = yaml.safe_load(yamlfile)["account_credentials"]
        for id, settings in exec_client_ids.items():
            main_venue, _, account_id = id.partition("-")
            creds = account_credentials.get(id, {})

            # For each exec_client_id generate SPOT and FUTURE exec_client separately
            for account_type in ("SPOT", "FUTURE"):
                client_id = f"{main_venue}_{account_type}-{account_id}"
                override_settings = {"defaultType": account_type.lower(), **creds, **settings}
                if client_id in exec_clients:
                    for key, val in override_settings.items():
                        if key not in exec_clients[client_id]:
                            exec_clients[client_id][key] = val
                else:
                    exec_clients[client_id] = override_settings

    node_cfg["exec_clients"] = exec_clients




from nautilus_trader.model.data.tick cimport Tick
from nautilus_trader.model.identifiers cimport InstrumentId
from nautilus_trader.model.objects cimport Price


cdef class MarkTick(Tick):
    cdef readonly Price price
    """The events close price.\n\n:returns: `Price`"""

    @staticmethod
    cdef MarkTick from_dict_c(dict values)

    @staticmethod
    cdef dict to_dict_c(MarkTick obj)




from libc.stdint cimport int64_t
from nautilus_trader.core.uuid cimport UUID4
from nautilus_trader.execution.reports cimport PositionStatusReport as NautilusPositionStatusReport
from nautilus_trader.model.c_enums.position_side cimport PositionSide
from nautilus_trader.model.identifiers cimport AccountId
from nautilus_trader.model.identifiers cimport ClientOrderId
from nautilus_trader.model.identifiers cimport InstrumentId
from nautilus_trader.model.identifiers cimport PositionId
from nautilus_trader.model.identifiers cimport TradeId
from nautilus_trader.model.identifiers cimport Venue
from nautilus_trader.model.identifiers cimport VenueOrderId
from nautilus_trader.model.objects cimport Price
from nautilus_trader.model.objects cimport Quantity


cdef class PositionStatusReport(NautilusPositionStatusReport):
    def __init__(
        self,
        AccountId account_id not None,
        InstrumentId instrument_id not None,
        PositionSide position_side,
        Quantity quantity not None,
        UUID4 report_id not None,
        int64_t ts_last,
        int64_t ts_init,
        PositionId venue_position_id = None,  # Can be None
        Price entry_px = None,
        Price last_px = None,
    ):
        super().__init__(
            account_id,
            instrument_id,
            position_side,
            quantity,
            report_id,
            ts_last,
            ts_init,
            venue_position_id,
        )

        self.entry_px = entry_px
        self.last_px = last_px



from nautilus_trader.serialization.base cimport Serializer

from nacre.actors.pubsub cimport PubSub
from nacre.serialization.base cimport AsyncSerializer


cdef class KafkaPubSub(PubSub):
    cdef bytes _key
    cdef object _producer
    cdef object _consumer
    cdef object config
    # cdef Serializer _serializer
    cdef AsyncSerializer _serializer
    cdef object _start_producer_task
    cdef object _stop_producer_task

    cpdef str detect_topic(self, object obj)



class ZbError(Exception):
    """
    The base class for all `Zb` specific errors.
    """

    def __init__(self, status, message):
        self.status = status
        self.message = message

    def __repr__(self) -> str:
        return f"{type(self).__name__}(" f"status={self.status}, " f"message={self.message})"


class ZbServerError(ZbError):
    """
    Represents a `Zb` specific 500 series HTTP error.
    """

    def __init__(self, status, message, headers):
        self.status = status
        self.message = message
        self.headers = headers


class ZbClientError(ZbError):
    """
    Represents a `Zb` specific 400 series HTTP error.
    """

    def __init__(self, status, message, headers):
        self.status = status
        self.message = message
        self.headers = headers


class ZbOperationError(ZbError):
    """
    Represents a `Zb` application logical error.
    """

    def __init__(self, status, message):
        self.status = status
        self.message = message


class ZbSpotOperationError(ZbError):
    """
    Represents a `Zb` application logical error.
    """

    def __init__(self, status, message):
        self.status = status
        self.message = message



from typing import Optional

from nautilus_trader.config.live import LiveDataClientConfig
from nautilus_trader.config.live import LiveExecClientConfig
from nautilus_trader.model.enums import OMSType

from nacre.adapters.zb.common.enums import ZbAccountType


class ZbDataClientConfig(LiveDataClientConfig):
    """
    Configuration for ``ZbDataClient`` instances.
    Parameters
    ----------
    api_key : str, optional
        The Zb API public key.
    api_secret : str, optional
        The Zb API private key.
    account_type : ZbAccountType, default ZbAccountType.SPOT
        The account type for the client.
    base_url_http : str, optional
        The HTTP client custom endpoint override.
    base_url_ws : str, optional
        The WebSocket client custom endpoint override.
    http_proxy : str, optional
        The HTTP proxy url.
    """

    api_key: Optional[str] = None
    api_secret: Optional[str] = None
    account_type: ZbAccountType = ZbAccountType.SPOT
    base_url_http: Optional[str] = None
    base_url_ws: Optional[str] = None
    http_proxy: Optional[str] = None


class ZbExecClientConfig(LiveExecClientConfig):
    """
    Configuration for ``ZbExecutionClient`` instances.
    Parameters
    ----------
    api_key : str, optional
        The Zb API public key.
    api_secret : str, optional
        The Zb API private key.
    oms_type: OMSType, default OMSType.NETTING
        OMSType for ZB, NETTING on spot, NETTING/HEDGING on futures
    account_type : ZbAccountType, default ZbAccountType.SPOT
        The account type for the client.
    base_url_http : str, optional
        The HTTP client custom endpoint override.
    base_url_ws : str, optional
        The WebSocket client custom endpoint override.
    http_proxy : str, optional
        The HTTP proxy url.
    """

    api_key: Optional[str] = None
    api_secret: Optional[str] = None
    oms_type: OMSType = OMSType.NETTING
    account_type: ZbAccountType = ZbAccountType.SPOT
    base_url_http: Optional[str] = None
    base_url_ws: Optional[str] = None
    http_proxy: Optional[str] = None



import asyncio
import hashlib
import hmac
import json
import time
from typing import Any, Callable, Dict, Optional  # noqa: TYP001

from nautilus_trader.common.clock import LiveClock
from nautilus_trader.common.logging import Logger

from nacre.adapters.zb.common import format_market
from nacre.adapters.zb.common import format_websocket_market
from nacre.adapters.zb.websocket.client import ZbWebSocketClient


class ZbSpotWebSocket(ZbWebSocketClient):
    """
    Provides access to the `Zb SPOT` streaming WebSocket API.
    """

    BASE_URL = "wss://api.zb.com/websocket"

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        clock: LiveClock,
        logger: Logger,
        handler: Callable[[bytes], None],
        base_url: Optional[str] = None,
    ):
        super().__init__(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=handler,
            base_url=base_url or self.BASE_URL,
        )

    async def ping(self):
        if not self.retrying:
            await self.send("ping".encode())

    async def _subscribe_channel(self, channel: str, **kwargs):
        kwargs["event"] = "addChannel"

        await super()._subscribe_channel(channel, **kwargs)

    async def subscribe_markets(self):
        await self._subscribe_channel(channel="markets")

    async def subscribe_ticker(self, symbol: str):
        channel = f"{format_websocket_market(symbol)}_ticker"
        await self._subscribe_channel(channel=channel)

    async def subscribe_depth(self, symbol: str):
        channel = f"{format_websocket_market(symbol)}_depth"
        await self._subscribe_channel(channel=channel)

    async def subscribe_trades(self, symbol: str):
        channel = f"{format_websocket_market(symbol)}_trades"
        await self._subscribe_channel(channel=channel)

    async def subscribe_quick_depth(self, symbol: str):
        channel = f"{format_websocket_market(symbol)}_quick_depth"
        await self._subscribe_channel(channel=channel)


class ZbSpotUserDataWebSocket(ZbWebSocketClient):
    BASE_URL = "wss://api.zb.com/websocket"

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        clock: LiveClock,
        logger: Logger,
        handler: Callable[[bytes], None],
        key: str,
        hashed_secret: str,
        base_url: Optional[str] = None,
    ):
        super().__init__(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=handler,
            base_url=base_url or self.BASE_URL,
        )

        self._key = key
        self._hashed_secret = hashed_secret
        # self.is_logged_in = False

    async def ping(self):
        if not self.retrying:
            try:
                await self.send("ping".encode())
            except ConnectionResetError:
                return

    # async def on_post_connect(self):
    #     await super().on_post_connect()
    #     # self.is_logged_in = True

    async def _request_channel(self, channel: str, **kwargs):
        kwargs["event"] = "addChannel"
        kwargs["accesskey"] = self._key
        kwargs["sign"] = self._get_sign(channel, kwargs)
        await super()._request_channel(channel, **kwargs)

    async def _subscribe_channel(self, channel: str, **kwargs):
        kwargs["event"] = "addChannel"
        kwargs["accesskey"] = self._key
        kwargs["sign"] = self._get_sign(channel, kwargs)
        await super()._subscribe_channel(channel, **kwargs)

    # async def logged_in(self):
    #     while not self.is_logged_in:
    #         await self._sleep0()
    #     self._log.debug("Websocket logged in")

    def _get_sign(self, channel: str, payload: Dict[str, Any]) -> str:
        params = {"channel": channel, **payload}
        sorted_params = dict(sorted(params.items()))
        query_string = json.dumps(sorted_params, separators=(",", ":"))
        return hmac.new(
            bytes(self._hashed_secret, encoding="utf-8"), query_string.encode("utf-8"), hashlib.md5
        ).hexdigest()

    async def subscribe_recent_order(self, market: str) -> None:
        payload = {
            "market": f"{format_market(market)}",
        }
        await self._subscribe_channel(channel="push_user_record", **payload)

    async def subscribe_order_update(self, market: str) -> None:
        payload = {
            "market": f"{format_market(market)}",
        }
        await self._subscribe_channel(channel="push_user_incr_record", **payload)

    async def subscribe_asset_snapshot(self):
        await self._subscribe_channel(channel="push_user_asset")

    async def subscribe_asset_update(self):
        await self._subscribe_channel(channel="push_user_incr_asset")

    async def get_account_info(self):
        payload = {"no": str(int(time.time() * 1000))}
        await self._request_channel(channel="getaccountinfo", **payload)



import asyncio
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, List, Optional

import orjson
import pandas as pd
from nautilus_trader.cache.cache import Cache
from nautilus_trader.common.clock import LiveClock
from nautilus_trader.common.logging import LogColor
from nautilus_trader.common.logging import Logger
from nautilus_trader.core.datetime import millis_to_nanos
from nautilus_trader.execution.messages import CancelAllOrders
from nautilus_trader.execution.messages import CancelOrder
from nautilus_trader.execution.messages import ModifyOrder
from nautilus_trader.execution.messages import SubmitOrder
from nautilus_trader.execution.messages import SubmitOrderList
from nautilus_trader.execution.reports import OrderStatusReport
from nautilus_trader.execution.reports import PositionStatusReport
from nautilus_trader.execution.reports import TradeReport
from nautilus_trader.model.enums import AccountType
from nautilus_trader.model.enums import LiquiditySide
from nautilus_trader.model.enums import OMSType
from nautilus_trader.model.enums import OrderSide
from nautilus_trader.model.enums import OrderStatus
from nautilus_trader.model.enums import OrderType
from nautilus_trader.model.enums import PositionSide
from nautilus_trader.model.enums import TimeInForce
from nautilus_trader.model.enums import TimeInForceParser
from nautilus_trader.model.identifiers import AccountId
from nautilus_trader.model.identifiers import ClientId
from nautilus_trader.model.identifiers import ClientOrderId
from nautilus_trader.model.identifiers import InstrumentId
from nautilus_trader.model.identifiers import TradeId
from nautilus_trader.model.identifiers import Venue
from nautilus_trader.model.identifiers import VenueOrderId
from nautilus_trader.model.instruments.base import Instrument
from nautilus_trader.model.objects import AccountBalance
from nautilus_trader.model.objects import MarginBalance
from nautilus_trader.model.objects import Money
from nautilus_trader.model.objects import Price
from nautilus_trader.model.objects import Quantity
from nautilus_trader.model.orders.base import Order
from nautilus_trader.model.orders.list import OrderList
from nautilus_trader.msgbus.bus import MessageBus

from nacre.adapters.zb.http.api.future_account import ZbFutureAccountHttpAPI
from nacre.adapters.zb.http.api.future_market import ZbFutureMarketHttpAPI
from nacre.adapters.zb.http.client import ZbHttpClient
from nacre.adapters.zb.http.error import ZbError
from nacre.adapters.zb.parsing import parse_account_balances_raw
from nacre.adapters.zb.parsing import parse_account_margins_raw
from nacre.adapters.zb.parsing import parse_future_order_status
from nacre.adapters.zb.parsing import parse_order_fill
from nacre.adapters.zb.parsing import parse_position
from nacre.adapters.zb.parsing import parse_zb_order_side
from nacre.adapters.zb.parsing import parse_zb_order_type
from nacre.adapters.zb.parsing import zb_order_params
from nacre.adapters.zb.providers import ZbInstrumentProvider
from nacre.adapters.zb.websocket.futures import ZbFuturesUserDataWebSocketClient
from nacre.live.execution_client import LiveExecutionClient


VALID_TIF = (TimeInForce.GTC, TimeInForce.FOK, TimeInForce.IOC)


class ZbFuturesExecutionClient(LiveExecutionClient):
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        client: ZbHttpClient,
        name: str,
        account_id: AccountId,
        venue: Venue,
        oms_type: OMSType,
        msgbus: MessageBus,
        cache: Cache,
        clock: LiveClock,
        logger: Logger,
        instrument_provider: ZbInstrumentProvider,
        base_url_ws: Optional[str] = None,
    ):
        # HTTP API
        self._account_api = ZbFutureAccountHttpAPI(client=client)
        self._market_api = ZbFutureMarketHttpAPI(client=client)

        # Websocket API
        self._ws_user_api = ZbFuturesUserDataWebSocketClient(
            loop=loop,
            clock=clock,
            logger=logger,
            handler=self._handle_user_ws_message,
            key=client.api_key,
            hashed_secret=client.hashed_secret,
            base_url=base_url_ws,
        )

        self._client = client
        self._proxy = client._proxy
        # Hot caches
        self._instrument_ids: Dict[str, InstrumentId] = {}
        self._order_types: Dict[VenueOrderId, OrderType] = {}
        self._last_filled: Dict[VenueOrderId, Quantity] = {}

        # cache for position update
        self._position_side_by_client_order_id: Dict[ClientOrderId, PositionSide] = {}

        super().__init__(
            loop=loop,
            client_id=ClientId(name),
            venue=venue,
            oms_type=oms_type,
            instrument_provider=instrument_provider,
            account_type=AccountType.MARGIN,
            base_currency=None,
            msgbus=msgbus,
            cache=cache,
            clock=clock,
            logger=logger,
            config={"name": f"ZbExecClient-{name}"},
        )

        self._set_account_id(account_id)
        self._set_venue(Venue(account_id.issuer))

    def connect(self) -> None:
        self._log.info("Connecting...")
        self._loop.create_task(self._connect())

    def disconnect(self) -> None:
        self._log.info("Disconnecting...")
        self._loop.create_task(self._disconnect())

    async def _connect(self) -> None:
        if not self._client.connected:
            await self._client.connect()
        try:
            await self._instrument_provider.initialize()
            response = await self._account_api.get_account(future_account_type=1)  # Ignore QC
        except Exception as ex:
            self._log.exception("Error on connect", ex)
            return

        balances = self.parse_account_balances(response)
        margins = self.parse_account_margins(response)
        self._update_account_state(balances, margins)

        # Connect WebSocket clients
        await self._connect_websockets()

        self._set_connected(True)
        assert self.is_connected
        self._log.info("Connected.")

    async def _connect_websockets(self) -> None:
        kwargs = {"proxy": self._proxy} if self._proxy else {}
        await self._ws_user_api.connect(**kwargs)
        await asyncio.sleep(2)  # Wait for login, since login method in separate write coroutine
        await asyncio.gather(
            self._ws_user_api.subscribe_asset_update(),
            self._ws_user_api.subscribe_order_update(),
            self._ws_user_api.subscribe_position_update(),
        )
        # await self._ws_user_api.logged_in()

    async def _disconnect(self) -> None:
        # Disconnect WebSocket clients
        if self._ws_user_api.is_connected:
            await self._ws_user_api.disconnect()

        if self._client.connected:
            await self._client.disconnect()

        self._set_connected(False)
        self._log.info("Disconnected.")

    def _update_account_state(
        self, balances: List[AccountBalance], margins: Optional[List[MarginBalance]] = None
    ) -> None:
        if not balances:
            return

        self.generate_account_state(
            balances=balances,
            margins=[] if margins is None else margins,
            reported=True,
            ts_event=self._clock.timestamp_ns(),  # zb account doesn't provide updateTime
        )

    async def _verify_order_status(self, order: Order):
        try:
            response = await self._account_api.get_order(
                symbol=order.instrument_id.symbol.value,
                client_order_id=order.client_order_id.value,
            )
            self._log.debug(f"Fetched order status: {response}")

            venue_order_id = VenueOrderId(response["id"])
            self.generate_order_accepted(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                venue_order_id=venue_order_id,
                ts_event=self._clock.timestamp_ns(),
            )
        except ZbError as ex:
            self.generate_order_rejected(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                reason=ex.message,
                ts_event=self._clock.timestamp_ns(),
            )

    # -- EXECUTION REPORTS -------------------------------------------------------------------------

    async def generate_order_status_report(
        self,
        venue_order_id: VenueOrderId = None,
    ) -> Optional[OrderStatusReport]:
        if venue_order_id is None:
            return None

        return None

        # try:
        #     response = await self._account_api.get_order(order_id=venue_order_id.value)
        # except ZbError as ex:
        #     self._log.error(ex.message)  # type: ignore  # TODO(cs): Improve errors
        #     return None
        #
        # # Get instrument
        # instrument_id: InstrumentId = self._get_cached_instrument_id(response)
        # instrument = self._instrument_provider.find(instrument_id)
        # if instrument is None:
        #     self._log.error(
        #         f"Cannot generate order status report: "
        #         f"no instrument found for {instrument_id}.",
        #     )
        #     return None
        #
        # return self.parse_order_status(
        #     account_id=self.account_id,
        #     instrument=instrument,
        #     data=response,
        #     report_id=self._uuid_factory.generate(),
        #     ts_init=self._clock.timestamp_ns(),
        # )

    async def generate_order_status_reports(
        self,
        instrument_id: InstrumentId = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        open_only: bool = False,
    ) -> List[OrderStatusReport]:
        self._log.info(f"Generating OrderStatusReports for {self.id}...")

        reports: List[OrderStatusReport] = []
        reports += await self._get_order_status_reports(
            instrument_id=instrument_id,
            start=start,
            end=end,
            open_only=open_only,
        )

        len_reports = len(reports)
        plural = "" if len_reports == 1 else "s"
        self._log.info(f"Generated {len(reports)} OrderStatusReport{plural}.")

        return reports

    async def generate_trade_reports(
        self,
        instrument_id: InstrumentId = None,
        venue_order_id: VenueOrderId = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> List[TradeReport]:
        self._log.info(f"Generating TradeReports for {self.id}...")

        # reports: List[TradeReport] = []

        reports = await self._get_trade_reports(
            instrument_id=instrument_id,
            venue_order_id=venue_order_id,
            start=start,
            end=end,
        )

        reports = sorted(reports, key=lambda x: x.trade_id)

        len_reports = len(reports)
        plural = "" if len_reports == 1 else "s"
        self._log.info(f"Generated {len(reports)} TradeReport{plural}.")

        return reports

    async def generate_position_status_reports(
        self,
        instrument_id: InstrumentId = None,
        start: datetime = None,
        end: datetime = None,
    ) -> List[PositionStatusReport]:

        self._log.info(f"Generating PositionStatusReports for {self.id}...")
        return []

    def _handle_user_ws_message(self, raw: bytes):  # noqa: C901
        msg: Dict[str, Any] = orjson.loads(raw)
        data = msg.get("data")

        chan: Optional[str] = msg.get("channel")
        err: Optional[str] = msg.get("errorCode")

        if err:
            self._log.error(f"Subscribe failed for channel: {chan}, error: {err}")
            return
        if not chan:
            if msg.get("action") == "login" and data == "success":
                pass
            elif msg.get("action") == "pong":
                pass
            return

        if chan == "Fund.assetChange":
            self._handle_asset_update(data)
        elif chan == "Fund.balance":
            self._handle_asset_snapshot(data)
        elif chan == "Positions.change":
            self._handle_position_update(data)
        elif chan == "Trade.orderChange":
            self._handle_order_update(data)
        elif chan == "Trade.order":
            self._handle_order_submit(data)
        elif chan == "Trade.cancelOrder":
            self._handle_order_cancel(data)
        elif chan == "Trade.cancelAllOrders":
            self._handle_order_cancel_all(data)
        else:
            self._log.warning(f"Unrecognized websocket message type, msg {msg}")
            return

    def _handle_order_submit(self, data: Dict):
        venue_order_id = VenueOrderId(data["orderId"])
        client_order_id = ClientOrderId(data["orderCode"])

        order = self._cache.order(client_order_id)
        if order is None:
            self._log.error(
                f"Cannot handle order update: " f"order for {client_order_id} not found",
            )
            return

        self.generate_order_accepted(
            strategy_id=order.strategy_id,
            instrument_id=order.instrument_id,
            client_order_id=client_order_id,
            venue_order_id=venue_order_id,
            ts_event=self._clock.timestamp_ns(),
        )

    def _handle_order_cancel(self, data: str):
        venue_order_id = VenueOrderId(data)
        client_order_id = self._cache.client_order_id(venue_order_id)
        if client_order_id is None:
            return
        order = self._cache.order(client_order_id)
        self.generate_order_canceled(
            strategy_id=order.strategy_id,
            instrument_id=order.instrument_id,
            client_order_id=client_order_id,
            venue_order_id=venue_order_id,
            ts_event=self._clock.timestamp_ns(),
        )

    def _handle_order_cancel_all(self, data: List):
        pass

    def _handle_asset_snapshot(self, data: List[Dict]):
        balances = parse_account_balances_raw(self._instrument_provider, data)
        self._update_account_state(balances)

    def _handle_asset_update(self, data: Dict[str, Any]):
        pass

    def _handle_position_update(self, data: Dict[str, Any]):
        pass

    def _handle_order_update(self, data: Dict[str, Any]):  # noqa: C901
        self._log.debug(f"Received order update {data}")

        venue_order_id = VenueOrderId(data["id"])
        client_order_id_str = data.get("orderCode")
        if client_order_id_str is None:
            self._log.warning(f"Cannot fill un-cached order with {repr(venue_order_id)}")
            return
        client_order_id = ClientOrderId(client_order_id_str)

        strategy_id = self._cache.strategy_id_for_order(client_order_id)
        if strategy_id is None:
            self._log.error(
                f"Cannot handle order update: " f"strategy ID for {client_order_id} not found",
            )
            return
        ts_event: int = millis_to_nanos(int(data["modifyTime"]))

        instrument_id = self._instrument_provider.find_instrument_id_by_local_market_id(
            data["marketId"]
        )

        order_status = data["showStatus"]
        if order_status == 1:  # 1 for accepted
            self.generate_order_accepted(
                strategy_id=strategy_id,
                instrument_id=instrument_id,
                client_order_id=client_order_id,
                venue_order_id=venue_order_id,
                ts_event=ts_event,
            )
        elif order_status == 2 or order_status == 3:  # 2 for partial filled, 3 for filled
            instrument: Instrument = self._instrument_provider.find(instrument_id)
            order_type = parse_zb_order_type(data["action"])

            # calculate last qty
            current_filled = Quantity.from_str(data["tradeAmount"])
            if venue_order_id not in self._last_filled:
                self._last_filled[venue_order_id] = current_filled
                last_qty = current_filled
            else:
                last_qty = Quantity(
                    current_filled - self._last_filled[venue_order_id], instrument.size_precision
                )
                self._last_filled[venue_order_id] = current_filled

            # Zb future might push same event more than once
            if last_qty == 0:
                self._log.warning(f"ZbFuture pushed same {client_order_id} fill event again")
                return

            # FIX: get commission and liquidity_side, hard code for now
            if order_type == OrderType.MARKET:
                commission_fee = (
                    Decimal(data["avgPrice"]) * last_qty.as_decimal()
                ) * instrument.taker_fee
                liquidity_side = LiquiditySide.TAKER
            else:
                commission_fee = (
                    Decimal(data["avgPrice"]) * last_qty.as_decimal()
                ) * instrument.maker_fee
                liquidity_side = LiquiditySide.MAKER
            commission = Money(commission_fee, instrument.quote_currency)

            venue_position_id = None
            # TODO: Avoid same side order generate 2 different position_id
            if (
                self.oms_type == OMSType.HEDGING
                and client_order_id in self._position_side_by_client_order_id
            ):

                side = self._position_side_by_client_order_id[client_order_id]
                open_positions = self._cache.positions_open(
                    instrument_id=instrument_id, strategy_id=strategy_id
                )
                position = next(
                    filter(lambda position: position.side == side, open_positions), None
                )
                if position is not None:
                    venue_position_id = position.id

            self.generate_order_filled(
                strategy_id=strategy_id,
                instrument_id=instrument_id,
                client_order_id=client_order_id,
                venue_order_id=venue_order_id,
                venue_position_id=venue_position_id,
                trade_id=TradeId(
                    self._uuid_factory.generate().value,
                ),  # no trade id returned, use uuid instead
                order_side=parse_zb_order_side(data["type"]),
                order_type=order_type,
                last_qty=last_qty,
                last_px=Price.from_str(data["avgPrice"]),
                quote_currency=instrument.quote_currency,
                commission=commission,  # see above fix
                liquidity_side=liquidity_side,  # see above fix
                ts_event=ts_event,
            )
        elif order_status == 4:  # 4 for canceling
            self.generate_order_pending_cancel(
                strategy_id=strategy_id,
                instrument_id=instrument_id,
                client_order_id=client_order_id,
                venue_order_id=venue_order_id,
                ts_event=ts_event,
            )
        elif (
            order_status == 5 or order_status == 7
        ):  # 5 for canceled, 7 for partial filled partial cancel
            # Some order canceled immediately without "showStatus" = 1 pushed
            if (
                order_status == 5
                and self._cache.order(client_order_id).status == OrderStatus.SUBMITTED
            ):
                self.generate_order_accepted(
                    strategy_id=strategy_id,
                    instrument_id=instrument_id,
                    client_order_id=client_order_id,
                    venue_order_id=venue_order_id,
                    ts_event=ts_event,
                )
            self.generate_order_canceled(
                strategy_id=strategy_id,
                instrument_id=instrument_id,
                client_order_id=client_order_id,
                venue_order_id=venue_order_id,
                ts_event=ts_event,
            )
        elif order_status == 6:  # 6 for cancel failed
            self.generate_order_cancel_rejected(
                strategy_id=strategy_id,
                instrument_id=instrument_id,
                client_order_id=client_order_id,
                venue_order_id=venue_order_id,
                reason="",  # type: ignore  # TODO(cs): Improve errors
                ts_event=ts_event,
            )

    # -- COMMAND HANDLERS --------------------------------------------------------------------------

    def submit_order(self, command: SubmitOrder) -> None:
        order: Order = command.order
        if order.type == OrderType.STOP_MARKET:
            self._log.error(
                "Cannot submit order: "
                "STOP_MARKET orders not supported by the exchange for zb FUTURE markets. "
                "Use any of MARKET, LIMIT, STOP_LIMIT."
            )
            return
        elif order.type == OrderType.STOP_LIMIT:
            self._log.error(
                "Cannot submit order: "
                "STOP_LIMIT orders not supported by the exchange for zb FUTURE markets. "
                "Use any of MARKET, LIMIT, STOP_LIMIT."
            )
            return
        if order.time_in_force not in VALID_TIF:
            self._log.error(
                f"Cannot submit order: "
                f"{TimeInForceParser.to_str_py(order.time_in_force)} "
                f"not supported by the exchange. Use any of {VALID_TIF}.",
            )
            return
        self._loop.create_task(self._submit_order(command))

    def submit_order_list(self, command: SubmitOrderList) -> None:
        order_list: OrderList = command.list
        self._loop.create_task(self._submit_order_list(order_list))

    def modify_order(self, command: ModifyOrder) -> None:
        self._log.error(  # pragma: no cover
            "Cannot modify order: Not supported by the exchange.",
        )

    def cancel_order(self, command: CancelOrder) -> None:
        self._loop.create_task(self._cancel_order(command))

    def cancel_all_orders(self, command: CancelAllOrders) -> None:
        self._loop.create_task(self._cancel_all_order(command))

    async def _submit_order(self, command: SubmitOrder) -> None:
        order: Order = command.order
        self._log.debug(f"Submitting {order}.")

        position = None
        position_id = command.position_id
        if position_id is not None:
            position = self._cache.position(position_id)
            if position is None:
                self._log.error(f"Position {position_id} not found")
                return
        elif (
            self.oms_type == OMSType.HEDGING
        ):  # When not specify position_id, open position with same side
            if order.side == OrderSide.BUY:
                side = PositionSide.LONG
            else:
                side = PositionSide.SHORT
            self._position_side_by_client_order_id[order.client_order_id] = side

        # Generate event here to ensure correct ordering of events
        self.generate_order_submitted(
            strategy_id=command.strategy_id,
            instrument_id=command.instrument_id,
            client_order_id=order.client_order_id,
            ts_event=self._clock.timestamp_ns(),
        )

        try:
            kwargs = zb_order_params(order, self.oms_type, position)
            response = await self._account_api.new_order(**kwargs)
            self._log.debug(f"Submit order response {response}")
            # Submit order through REST & Websocket same time
            await self._ws_user_api.new_order(**kwargs)
        except ZbError as ex:
            self.generate_order_rejected(
                strategy_id=command.strategy_id,
                instrument_id=command.instrument_id,
                client_order_id=order.client_order_id,
                reason=ex.message,
                ts_event=self._clock.timestamp_ns(),  # TODO(cs): Parse from response
            )
        except Exception as ex:
            self._log.error(f"Error occurs in submit order: {ex}, verifying order ...")
            await self._verify_order_status(order)

    async def _submit_order_list(self, order_list: OrderList) -> None:
        batch_params = []
        for order in order_list.orders:
            if order.contingency_ids:  # TODO(cs): Implement
                self._log.warning(f"Cannot yet handle contingency orders, {order}.")
            params = zb_order_params(order, self.oms_type)
            params["orderCode"] = params.pop("client_order_id")
            batch_params.append(params)

        response = await self._account_api.new_batch_order(orders=batch_params)
        self._log.debug(f"Submit batch order response {response}", LogColor.YELLOW)

    async def _cancel_order(self, order: CancelOrder) -> None:
        self._log.debug(f"Canceling order {order.client_order_id.value}.")

        self.generate_order_pending_cancel(
            strategy_id=order.strategy_id,
            instrument_id=order.instrument_id,
            client_order_id=order.client_order_id,
            venue_order_id=order.venue_order_id,
            ts_event=self._clock.timestamp_ns(),
        )

        try:
            if order.venue_order_id is not None:
                response = await self._account_api.cancel_order(
                    symbol=order.instrument_id.symbol.value,
                    order_id=order.venue_order_id.value,
                )
            else:
                response = await self._account_api.cancel_order(
                    symbol=order.instrument_id.symbol.value,
                    client_order_id=order.client_order_id.value,
                )
            # Cancel order through REST & Websocket same time
            await self._ws_user_api.cancel_order(
                symbol=order.instrument_id.symbol.value,
                client_order_id=order.client_order_id.value,
            )
            self._log.debug(f"Cancel order response {response}")
        except Exception as ex:
            if isinstance(ex, ZbError):
                message = ex.message
            else:
                message = repr(ex)
                self._log.exception("Error on cancel order", ex)
            self.generate_order_cancel_rejected(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                venue_order_id=order.venue_order_id,
                reason=message,  # type: ignore  # TODO(cs): Improve errors
                ts_event=self._clock.timestamp_ns(),  # TODO(cs): Parse from response
            )

    async def _cancel_all_order(self, command: CancelAllOrders) -> None:
        self._log.debug(f"Canceling all orders for {command.instrument_id.value}.")

        # Cancel all in-flight orders
        inflight_orders = self._cache.orders_inflight(
            instrument_id=command.instrument_id,
            strategy_id=command.strategy_id,
        )
        for order in inflight_orders:
            self.generate_order_pending_cancel(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                venue_order_id=order.venue_order_id,
                ts_event=self._clock.timestamp_ns(),
            )

        # Cancel all working orders
        open_orders = self._cache.orders_open(
            instrument_id=command.instrument_id,
            strategy_id=command.strategy_id,
        )

        for order in open_orders:
            self.generate_order_pending_cancel(
                strategy_id=order.strategy_id,
                instrument_id=order.instrument_id,
                client_order_id=order.client_order_id,
                venue_order_id=order.venue_order_id,
                ts_event=self._clock.timestamp_ns(),
            )
        try:
            response = await self._account_api.cancel_open_orders(
                command.instrument_id.symbol.value
            )
            # # Cancel order through REST & Websocket same time
            # await self._ws_user_api.cancel_open_orders(command.instrument_id.symbol.value)
            self._log.debug(f"Cancel all order response {response}")
        except ZbError as ex:
            self._log.error(ex)  # type: ignore  # TODO(cs): Improve errors
        except Exception as ex:
            self._log.exception("Error on cancel all orders", ex)

    async def _get_order_status_reports(
        self,
        instrument_id: InstrumentId = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        open_only: bool = False,
    ) -> List[OrderStatusReport]:
        reports: List[OrderStatusReport] = []

        if instrument_id is None:
            self._log.warning("InstrumentId not specified, zb does not support all orders query")
            return reports

        try:
            if open_only:
                response: Dict[str, Any] = await self._account_api.get_open_orders(
                    symbol=instrument_id.symbol.value,
                )
            else:
                response = await self._account_api.get_all_orders(
                    symbol=instrument_id.symbol.value,
                )
        except ZbError as ex:
            self._log.error(ex.message)  # type: ignore  # TODO(cs): Improve errors
            return []

        if response:
            for data in response["data"]["list"]:
                created_at = pd.to_datetime(data["createTime"], unit="ms")
                if start is not None and created_at < start:
                    continue
                if end is not None and created_at > end:
                    continue

                # Get instrument
                instrument_id = instrument_id or self._get_cached_instrument_id(data)
                instrument = self._instrument_provider.find(instrument_id)
                if instrument is None:
                    self._log.error(
                        f"Cannot generate order status report: "
                        f"no instrument found for {instrument_id}.",
                    )
                    continue

                report: OrderStatusReport = self.parse_order_status(
                    account_id=self.account_id,
                    instrument=instrument,
                    data=data,
                    report_id=self._uuid_factory.generate(),
                    ts_init=self._clock.timestamp_ns(),
                )

                self._log.debug(f"Received {report}.")
                reports.append(report)

        return reports

    async def _get_trade_reports(
        self,
        instrument_id: InstrumentId = None,
        venue_order_id: VenueOrderId = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> List[TradeReport]:
        reports: List[TradeReport] = []
        if instrument_id is None:
            return reports

        try:
            response: Dict[str, Any] = await self._account_api.get_history_trades(
                symbol=instrument_id.symbol.value,
                start_time=int(start.timestamp() * 1000) if start is not None else None,
                end_time=int(end.timestamp() * 1000) if end is not None else None,
            )
        except ZbError as ex:
            self._log.error(ex.message)  # type: ignore  # TODO(cs): Improve errors
            return []

        if response:
            for data in response["data"]["list"]:
                # Apply filter (FTX filters not working)
                created_at = pd.to_datetime(data["createTime"], unit="ms")
                if start is not None and created_at < start:
                    continue
                if end is not None and created_at > end:
                    continue

                # Get instrument
                instrument_id = instrument_id or self._get_cached_instrument_id(data)
                instrument = self._instrument_provider.find(instrument_id)
                if instrument is None:
                    self._log.error(
                        f"Cannot generate order status report: "
                        f"no instrument found for {instrument_id}.",
                    )
                    continue

                report: TradeReport = parse_order_fill(
                    account_id=self.account_id,
                    instrument=instrument,
                    data=data,
                    report_id=self._uuid_factory.generate(),
                    ts_init=self._clock.timestamp_ns(),
                )

                self._log.debug(f"Received {report}.")
                reports.append(report)

        return reports

    async def _get_position_status_report(
        self,
        instrument_id: InstrumentId = None,
        start: datetime = None,
        end: datetime = None,
    ) -> List[PositionStatusReport]:
        try:
            response: Dict[str, Any] = await self._account_api.get_positions()
        except ZbError as ex:
            self._log.error(ex.message)  # type: ignore  # TODO(cs): Improve errors
            return []

        reports = []
        if response:
            for data in response["data"]:
                # Get instrument
                instrument_id = instrument_id or self._get_cached_instrument_id(data)
                instrument = self._instrument_provider.find(instrument_id)
                if instrument is None:
                    self._log.error(
                        f"Cannot generate order status report: "
                        f"no instrument found for {instrument_id}.",
                    )
                    continue

                report: PositionStatusReport = parse_position(
                    account_id=self.account_id,
                    instrument=instrument,
                    data=data,
                    report_id=self._uuid_factory.generate(),
                    ts_init=self._clock.timestamp_ns(),
                )

                if report.quantity == 0:
                    continue  # Flat position
                self._log.debug(f"Received {report}.")
                reports.append(report)
        return reports

    # -- Others -------------------------------------------------------------------------

    def parse_account_balances(self, response: Dict) -> List[AccountBalance]:
        return parse_account_balances_raw(self._instrument_provider, response["data"]["assets"])

    def parse_account_margins(self, response: Dict) -> List[MarginBalance]:
        return parse_account_margins_raw(self._instrument_provider, response["data"]["accounts"])

    def parse_order_status(self, account_id, instrument, data, report_id, ts_init):
        return parse_future_order_status(account_id, instrument, data["data"], report_id, ts_init)

    def _get_cached_instrument_id(self, response: Dict) -> InstrumentId:
        return self._instrument_provider.find_instrument_id_by_local_market_id(response["marketId"])



import asyncio

import orjson
import pydantic

from libc.stdint cimport int64_t

from json.decoder import JSONDecodeError
from typing import Dict, Optional, Union

from aiohttp import web
from aiohttp.web_exceptions import HTTPBadRequest
from prometheus_async import aio

from nautilus_trader.common.actor cimport Actor
from nautilus_trader.common.logging cimport LogColor
from nautilus_trader.common.logging cimport Logger
from nautilus_trader.common.logging cimport LoggerAdapter
from nautilus_trader.common.logging cimport LogLevel
from nautilus_trader.core.correctness cimport Condition
from nautilus_trader.model.data.base cimport DataType
from nautilus_trader.model.data.base cimport GenericData
from nautilus_trader.model.data.tick cimport QuoteTick
from nautilus_trader.model.data.tick cimport TradeTick
from nautilus_trader.model.objects cimport BaseDecimal

from nacre.config import ExposerConfig
from nacre.model.data.tick cimport MarkTick


def default(obj):
    if isinstance(obj, BaseDecimal):
        return str(obj.as_decimal())
    raise TypeError

cdef class AccessLoggerAdapter(LoggerAdapter):
    cpdef void info(
        self, str msg,
        LogColor color=LogColor.NORMAL,
        dict extra=None,
    ) except *:
        """
        Log the given information message with the logger.

        Parameters
        ----------
        msg : str
            The message to log.
        color : LogColor, optional
            The color for the log record.
        extra : dict[str, object], optional
            The annotations for the log record.

        """
        Condition.not_none(msg, "msg")

        if self.is_bypassed:
            return

        cdef dict record = self._logger.create_record(
            level=LogLevel.DEBUG,  # Hard code for now, metrics log might be too much for INFO level
            color=color,
            component=self.component,
            msg=msg,
            annotations=extra,
        )

        self._logger.log_c(record)



cdef class Exposer(Actor):
    """
    Expose nacre internal states via http endpoint

    Parameters
    ----------
    config : ExposerConfig
        The actor configuration.

    Raises
    ------
    TypeError
        If `config` is not of type `ExposerConfig`.
    """
    def __init__(self, config: ExposerConfig):
        super().__init__(config)

        self._run_http_server_task = None
        self._runner = None
        self._loop = asyncio.get_event_loop()

        self.trader = None  # Initialized when registered

    cpdef void register_trader(self, Trader trader) except *:
        Condition.not_none(trader, "trader")
        self.trader = trader

    cpdef void on_start(self) except *:
        self._run_http_server_task = self._loop.create_task(self._run_web_server())
        self._log.info(f"Scheduled {self._run_http_server_task}")

    cpdef void on_stop(self) except *:
        self._loop.create_task(self._stop_server())

    async def _run_web_server(self):
        self._log.info(f"HTTP server starting...")

        app = web.Application()
        app.add_routes([
            # General
            web.get('/health', self.expose_health),
            web.get('/metrics', aio.web.server_stats),
        ])

        access_log = AccessLoggerAdapter(
            component_name=self.type.__name__,
            logger=self._log.get_logger(),
        )

        self._runner = web.AppRunner(
            app,
            access_log=access_log,
        )

        await self._runner.setup()
        site = web.TCPSite(
            self._runner,
            self._config.get("host"),
            self._config.get("port"),
        )

        await site.start()
        self._log.info(f"HTTP server started")

    async def _stop_server(self):
        if self._runner:
            await self._runner.shutdown()
        self._log.info(f"HTTP server stopped")

    async def expose_health(self, request):
        return web.json_response({"status": "OK"})
